import math
from typing import Union


def convert_rotation_to_matrix(rotation: Union[float, int]):
    rad = math.radians(rotation)
    return [math.cos(rad), -math.sin(rad), math.sin(rad), math.cos(rad)]


def convertRotationToMatrix(degrees: Union[float, int]):
    # first convert degrees to radians
    radians = degrees * math.pi / 180
    # now we do this:
    """
    [cos(r), -sin(r), 0]
    [sin(r), cos(r), 0]

    in this order:
    [cos(r), sin(r), -sin(r), cos(r), 0, 0]
    """
    cos = math.cos(radians)
    sin = math.sin(radians)
    return [cos, sin, -sin, cos, 0, 0]


if __name__ == "__main__":
    pass



from __future__ import annotations

from typing import TYPE_CHECKING

from ..constants import (
    DOMAIN,
    BOOKMARK_FOLDER_TIMELINE_FEATURES,
    COMMUNITY_NOTE_FEATURES,
    COMMUNITY_TWEETS_FEATURES,
    FEATURES,
    JOIN_COMMUNITY_FEATURES,
    LIST_FEATURES,
    NOTE_TWEET_FEATURES,
    SIMILAR_POSTS_FEATURES,
    TWEET_RESULT_BY_REST_ID_FEATURES,
    TWEET_RESULTS_BY_REST_IDS_FEATURES,
    USER_FEATURES,
    USER_HIGHLIGHTS_TWEETS_FEATURES
)
from ..utils import flatten_params, get_query_id

if TYPE_CHECKING:
    from ..guest.client import GuestClient
    from .client import Client

    ClientType = Client | GuestClient


class Endpoint:
    @staticmethod
    def url(path):
        return f'https://{DOMAIN}/i/api/graphql/{path}'

    SEARCH_TIMELINE = url('flaR-PUMshxFWZWPNpq4zA/SearchTimeline')
    SIMILAR_POSTS = url('EToazR74i0rJyZYalfVEAQ/SimilarPosts')
    CREATE_NOTE_TWEET = url('iCUB42lIfXf9qPKctjE5rQ/CreateNoteTweet')
    CREATE_TWEET = url('SiM_cAu83R0wnrpmKQQSEw/CreateTweet')
    CREATE_SCHEDULED_TWEET = url('LCVzRQGxOaGnOnYH01NQXg/CreateScheduledTweet')
    DELETE_TWEET = url('VaenaVgh5q5ih7kvyVjgtg/DeleteTweet')
    USER_BY_SCREEN_NAME = url('NimuplG1OB7Fd2btCLdBOw/UserByScreenName')
    USER_BY_REST_ID = url('tD8zKvQzwY3kdx5yz6YmOw/UserByRestId')
    TWEET_DETAIL = url('U0HTv-bAWTBYylwEMT7x5A/TweetDetail')
    TWEET_RESULT_BY_REST_ID = url('Xl5pC_lBk_gcO2ItU39DQw/TweetResultByRestId')
    FETCH_SCHEDULED_TWEETS = url('ITtjAzvlZni2wWXwf295Qg/FetchScheduledTweets')
    DELETE_SCHEDULED_TWEET = url('CTOVqej0JBXAZSwkp1US0g/DeleteScheduledTweet')
    RETWEETERS = url('X-XEqG5qHQSAwmvy00xfyQ/Retweeters')
    FAVORITERS = url('LLkw5EcVutJL6y-2gkz22A/Favoriters')
    FETCH_COMMUNITY_NOTE = url('fKWPPj271aTM-AB9Xp48IA/BirdwatchFetchOneNote')
    USER_TWEETS = url('QWF3SzpHmykQHsQMixG0cg/UserTweets')
    USER_TWEETS_AND_REPLIES = url('vMkJyzx1wdmvOeeNG0n6Wg/UserTweetsAndReplies')
    USER_MEDIA = url('2tLOJWwGuCTytDrGBg8VwQ/UserMedia')
    USER_LIKES = url('IohM3gxQHfvWePH5E3KuNA/Likes')
    USER_HIGHLIGHTS_TWEETS = url('tHFm_XZc_NNi-CfUThwbNw/UserHighlightsTweets')
    HOME_TIMELINE = url('-X_hcgQzmHGl29-UXxz4sw/HomeTimeline')
    HOME_LATEST_TIMELINE = url('U0cdisy7QFIoTfu3-Okw0A/HomeLatestTimeline')
    FAVORITE_TWEET = url('lI07N6Otwv1PhnEgXILM7A/FavoriteTweet')
    UNFAVORITE_TWEET = url('ZYKSe-w7KEslx3JhSIk5LA/UnfavoriteTweet')
    CREATE_RETWEET = url('ojPdsZsimiJrUGLR1sjUtA/CreateRetweet')
    DELETE_RETWEET = url('iQtK4dl5hBmXewYZuEOKVw/DeleteRetweet')
    CREATE_BOOKMARK = url('aoDbu3RHznuiSkQ9aNM67Q/CreateBookmark')
    BOOKMARK_TO_FOLDER = url('4KHZvvNbHNf07bsgnL9gWA/bookmarkTweetToFolder')
    DELETE_BOOKMARK = url('Wlmlj2-xzyS1GN3a6cj-mQ/DeleteBookmark')
    BOOKMARKS = url('qToeLeMs43Q8cr7tRYXmaQ/Bookmarks')
    BOOKMARK_FOLDER_TIMELINE = url('8HoabOvl7jl9IC1Aixj-vg/BookmarkFolderTimeline')
    BOOKMARKS_ALL_DELETE = url('skiACZKC1GDYli-M8RzEPQ/BookmarksAllDelete')
    BOOKMARK_FOLDERS_SLICE = url('i78YDd0Tza-dV4SYs58kRg/BookmarkFoldersSlice')
    EDIT_BOOKMARK_FOLDER = url('a6kPp1cS1Dgbsjhapz1PNw/EditBookmarkFolder')
    DELETE_BOOKMARK_FOLDER = url('2UTTsO-6zs93XqlEUZPsSg/DeleteBookmarkFolder')
    CREATE_BOOKMARK_FOLDER = url('6Xxqpq8TM_CREYiuof_h5w/createBookmarkFolder')
    FOLLOWERS = url('gC_lyAxZOptAMLCJX5UhWw/Followers')
    BLUE_VERIFIED_FOLLOWERS = url('VmIlPJNEDVQ29HfzIhV4mw/BlueVerifiedFollowers')
    FOLLOWERS_YOU_KNOW = url('f2tbuGNjfOE8mNUO5itMew/FollowersYouKnow')
    FOLLOWING = url('2vUj-_Ek-UmBVDNtd8OnQA/Following')
    USER_CREATOR_SUBSCRIPTIONS = url('Wsm5ZTCYtg2eH7mXAXPIgw/UserCreatorSubscriptions')
    USER_DM_REACTION_MUTATION_ADD_MUTATION = url('VyDyV9pC2oZEj6g52hgnhA/useDMReactionMutationAddMutation')
    USER_DM_REACTION_MUTATION_REMOVE_MUTATION = url('bV_Nim3RYHsaJwMkTXJ6ew/useDMReactionMutationRemoveMutation')
    DM_MESSAGE_DELETE_MUTATION = url('BJ6DtxA2llfjnRoRjaiIiw/DMMessageDeleteMutation')
    ADD_PARTICIPANTS_MUTATION = url('oBwyQ0_xVbAQ8FAyG0pCRA/AddParticipantsMutation')
    CREATE_LIST = url('EYg7JZU3A1eJ-wr2eygPHQ/CreateList')
    EDIT_LIST_BANNER = url('t_DsROHldculsB0B9BUAWw/EditListBanner')
    DELETE_LIST_BANNER = url('Y90WuxdWugtMRJhkXTdvzg/DeleteListBanner')
    UPDATE_LIST = url('dIEI1sbSAuZlxhE0ggrezA/UpdateList')
    LIST_ADD_MEMBER = url('lLNsL7mW6gSEQG6rXP7TNw/ListAddMember')
    LIST_REMOVE_MEMBER = url('cvDFkG5WjcXV0Qw5nfe1qQ/ListRemoveMember')
    LIST_MANAGEMENT_PACE_TIMELINE = url('47170qwZCt5aFo9cBwFoNA/ListsManagementPageTimeline')
    LIST_BY_REST_ID = url('9hbYpeVBMq8-yB8slayGWQ/ListByRestId')
    LIST_LATEST_TWEETS_TIMELINE = url('HjsWc-nwwHKYwHenbHm-tw/ListLatestTweetsTimeline')
    LIST_MEMBERS = url('BQp2IEYkgxuSxqbTAr1e1g/ListMembers')
    LIST_SUBSCRIBERS = url('74wGEkaBxrdoXakWTWMxRQ/ListSubscribers')
    SEARCH_COMMUNITY = url('daVUkhfHn7-Z8llpYVKJSw/CommunitiesSearchQuery')
    COMMUNITY_QUERY = url('lUBKrilodgg9Nikaw3cIiA/CommunityQuery')
    COMMUNITY_MEDIA_TIMELINE = url('Ht5K2ckaZYAOuRFmFfbHig/CommunityMediaTimeline')
    COMMUNITY_TWEETS_TIMELINE = url('mhwSsmub4JZgHcs0dtsjrw/CommunityTweetsTimeline')
    COMMUNITIES_MAIN_PAGE_TIMELINE = url('4-4iuIdaLPpmxKnA3mr2LA/CommunitiesMainPageTimeline')
    JOIN_COMMUNITY = url('xZQLbDwbI585YTG0QIpokw/JoinCommunity')
    LEAVE_COMMUNITY = url('OoS6Kd4-noNLXPZYHtygeA/LeaveCommunity')
    REQUEST_TO_JOIN_COMMUNITY = url('XwWChphD_6g7JnsFus2f2Q/RequestToJoinCommunity')
    MEMBERS_SLICE_TIMELINE_QUERY = url('KDAssJ5lafCy-asH4wm1dw/membersSliceTimeline_Query')
    MODERATORS_SLICE_TIMELINE_QUERY = url('9KI_r8e-tgp3--N5SZYVjg/moderatorsSliceTimeline_Query')
    COMMUNITY_TWEET_SEARCH_MODULE_QUERY = url('5341rmzzvdjqfmPKfoHUBw/CommunityTweetSearchModuleQuery')
    TWEET_RESULTS_BY_REST_IDS = url('PTN9HhBAlpoCTHfspDgqLA/TweetResultsByRestIds')


class GQLClient:
    def __init__(self, base: ClientType) -> None:
        self.base = base

    async def gql_get(
        self,
        url: str,
        variables: dict,
        features: dict | None = None,
        headers: dict | None = None,
        extra_params: dict | None = None,
        **kwargs
    ):
        params = {'variables': variables}
        if features is not None:
            params['features'] = features
        if extra_params is not None:
            params |= extra_params
        if headers is None:
            headers = self.base._base_headers
        return await self.base.get(url, params=flatten_params(params), headers=headers, **kwargs)

    async def gql_post(
        self,
        url: str,
        variables: dict,
        features: dict | None = None,
        headers: dict | None = None,
        extra_data: dict | None = None,
        **kwargs
    ):
        data = {'variables': variables, 'queryId': get_query_id(url)}
        if features is not None:
            data['features'] = features
        if extra_data is not None:
            data |= extra_data
        if headers is None:
            headers = self.base._base_headers
        return await self.base.post(url, json=data, headers=headers, **kwargs)

    async def search_timeline(
        self,
        query: str,
        product: str,
        count: int,
        cursor: str | None
    ):
        variables = {
            'rawQuery': query,
            'count': count,
            'querySource': 'typed_query',
            'product': product
        }
        if cursor is not None:
            variables['cursor'] = cursor
        return await self.gql_get(Endpoint.SEARCH_TIMELINE, variables, FEATURES)

    async def similar_posts(self, tweet_id: str):
        variables = {'tweet_id': tweet_id}
        return await self.gql_get(
            Endpoint.SIMILAR_POSTS,
            variables,
            SIMILAR_POSTS_FEATURES
        )

    async def create_tweet(
        self, is_note_tweet, text, media_entities,
        poll_uri, reply_to, attachment_url,
        community_id, share_with_followers,
        richtext_options, edit_tweet_id, limit_mode
    ):
        variables = {
            'tweet_text': text,
            'dark_request': False,
            'media': {
                'media_entities': media_entities,
                'possibly_sensitive': False
            },
            'semantic_annotation_ids': [],
        }

        if poll_uri is not None:
            variables['card_uri'] = poll_uri

        if reply_to is not None:
            variables['reply'] = {
                'in_reply_to_tweet_id': reply_to,
                'exclude_reply_user_ids': []
            }

        if limit_mode is not None:
            variables['conversation_control'] = {'mode': limit_mode}

        if attachment_url is not None:
            variables['attachment_url'] = attachment_url

        if community_id is not None:
            variables['semantic_annotation_ids'] = [{
                'entity_id': community_id,
                'group_id': '8',
                'domain_id': '31'
            }]
            variables['broadcast'] = share_with_followers

        if richtext_options is not None:
            is_note_tweet = True
            variables['richtext_options'] = {
                'richtext_tags': richtext_options
            }
        if edit_tweet_id is not None:
            variables['edit_options'] = {
                'previous_tweet_id': edit_tweet_id
            }

        if is_note_tweet:
            endpoint = Endpoint.CREATE_NOTE_TWEET
            features = NOTE_TWEET_FEATURES
        else:
            endpoint = Endpoint.CREATE_TWEET
            features = FEATURES
        return await self.gql_post(endpoint, variables, features)

    async def create_scheduled_tweet(self, scheduled_at, text, media_ids) -> str:
        variables = {
            'post_tweet_request': {
            'auto_populate_reply_metadata': False,
            'status': text,
            'exclude_reply_user_ids': [],
            'media_ids': media_ids
            },
            'execute_at': scheduled_at
        }
        return await self.gql_post(Endpoint.CREATE_SCHEDULED_TWEET, variables)

    async def delete_tweet(self, tweet_id):
        variables = {
            'tweet_id': tweet_id,
            'dark_request': False
        }
        return await self.gql_post(Endpoint.DELETE_TWEET, variables)

    async def user_by_screen_name(self, screen_name):
        variables = {
            'screen_name': screen_name,
            'withSafetyModeUserFields': False
        }
        params = {
            'fieldToggles': {'withAuxiliaryUserLabels': False}
        }
        return await self.gql_get(Endpoint.USER_BY_SCREEN_NAME, variables, USER_FEATURES, extra_params=params)

    async def user_by_rest_id(self, user_id):
        variables = {
            'userId': user_id,
            'withSafetyModeUserFields': True
        }
        return await self.gql_get(Endpoint.USER_BY_REST_ID, variables, USER_FEATURES)

    async def tweet_detail(self, tweet_id, cursor):
        variables = {
            'focalTweetId': tweet_id,
            'with_rux_injections': False,
            'includePromotedContent': True,
            'withCommunity': True,
            'withQuickPromoteEligibilityTweetFields': True,
            'withBirdwatchNotes': True,
            'withVoice': True,
            'withV2Timeline': True
        }
        if cursor is not None:
            variables['cursor'] = cursor
        params = {
            'fieldToggles': {'withAuxiliaryUserLabels': False}
        }
        return await self.gql_get(Endpoint.TWEET_DETAIL, variables, FEATURES, extra_params=params)

    async def fetch_scheduled_tweets(self):
        variables = {'ascending': True}
        return await self.gql_get(Endpoint.FETCH_SCHEDULED_TWEETS, variables)

    async def delete_scheduled_tweet(self, tweet_id):
        variables = {'scheduled_tweet_id': tweet_id}
        return await self.gql_post(Endpoint.DELETE_SCHEDULED_TWEET, variables)

    async def tweet_engagements(self, tweet_id, count, cursor, endpoint):
        variables = {
            'tweetId': tweet_id,
            'count': count,
            'includePromotedContent': True
        }
        if cursor is not None:
            variables['cursor'] = cursor
        return await self.gql_get(endpoint, variables, FEATURES)

    async def retweeters(self, tweet_id, count, cursor):
        return await self.tweet_engagements(tweet_id, count, cursor, Endpoint.RETWEETERS)

    async def favoriters(self, tweet_id, count, cursor):
        return await self.tweet_engagements(tweet_id, count, cursor, Endpoint.FAVORITERS)

    async def bird_watch_one_note(self, note_id):
        variables = {'note_id': note_id}
        return await self.gql_get(Endpoint.FETCH_COMMUNITY_NOTE, variables, COMMUNITY_NOTE_FEATURES)

    async def _get_user_tweets(self, user_id, count, cursor, endpoint):
        variables = {
            'userId': user_id,
            'count': count,
            'includePromotedContent': True,
            'withQuickPromoteEligibilityTweetFields': True,
            'withVoice': True,
            'withV2Timeline': True
        }
        if cursor is not None:
            variables['cursor'] = cursor
        return await self.gql_get(endpoint, variables, FEATURES)

    async def user_tweets(self, user_id, count, cursor):
        return await self._get_user_tweets(user_id, count, cursor, Endpoint.USER_TWEETS)

    async def user_tweets_and_replies(self, user_id, count, cursor):
        return await self._get_user_tweets(user_id, count, cursor, Endpoint.USER_TWEETS_AND_REPLIES)

    async def user_media(self, user_id, count, cursor):
        return await self._get_user_tweets(user_id, count, cursor, Endpoint.USER_MEDIA)

    async def user_likes(self, user_id, count, cursor):
        return await self._get_user_tweets(user_id, count, cursor, Endpoint.USER_LIKES)

    async def user_highlights_tweets(self, user_id, count, cursor):
        variables = {
            'userId': user_id,
            'count': count,
            'includePromotedContent': True,
            'withVoice': True
        }
        if cursor is not None:
            variables['cursor'] = cursor
        return await self.gql_get(
            Endpoint.USER_HIGHLIGHTS_TWEETS,
            variables,
            USER_HIGHLIGHTS_TWEETS_FEATURES,
            self.base._base_headers
        )

    async def home_timeline(self, count, seen_tweet_ids, cursor):
        variables = {
            'count': count,
            'includePromotedContent': True,
            'latestControlAvailable': True,
            'requestContext': 'launch',
            'withCommunity': True,
            'seenTweetIds': seen_tweet_ids or []
        }
        if cursor is not None:
            variables['cursor'] = cursor
        return await self.gql_post(Endpoint.HOME_TIMELINE, variables, FEATURES)

    async def home_latest_timeline(self, count, seen_tweet_ids, cursor):
        variables = {
            'count': count,
            'includePromotedContent': True,
            'latestControlAvailable': True,
            'requestContext': 'launch',
            'withCommunity': True,
            'seenTweetIds': seen_tweet_ids or []
        }
        if cursor is not None:
            variables['cursor'] = cursor
        return await self.gql_post(Endpoint.HOME_LATEST_TIMELINE, variables, FEATURES)

    async def favorite_tweet(self, tweet_id):
        variables = {'tweet_id': tweet_id}
        return await self.gql_post(Endpoint.FAVORITE_TWEET, variables)

    async def unfavorite_tweet(self, tweet_id):
        variables = {'tweet_id': tweet_id}
        return await self.gql_post(Endpoint.UNFAVORITE_TWEET, variables)

    async def retweet(self, tweet_id):
        variables = {'tweet_id': tweet_id, 'dark_request': False}
        return await self.gql_post(Endpoint.CREATE_RETWEET, variables)

    async def delete_retweet(self, tweet_id):
        variables = {'source_tweet_id': tweet_id,'dark_request': False}
        return await self.gql_post(Endpoint.DELETE_RETWEET, variables)

    async def create_bookmark(self, tweet_id):
        variables = {'tweet_id': tweet_id}
        return await self.gql_post(Endpoint.CREATE_BOOKMARK, variables)

    async def bookmark_tweet_to_folder(self, tweet_id, folder_id):
        variables = {
            'tweet_id': tweet_id,
            'bookmark_collection_id': folder_id
        }
        return await self.gql_post(Endpoint.BOOKMARK_TO_FOLDER, variables)

    async def delete_bookmark(self, tweet_id):
        variables = {'tweet_id': tweet_id}
        return await self.gql_post(Endpoint.DELETE_BOOKMARK, variables)

    async def bookmarks(self, count, cursor):
        variables = {
            'count': count,
            'includePromotedContent': True
        }
        features = FEATURES | {
            'graphql_timeline_v2_bookmark_timeline': True
        }
        if cursor is not None:
            variables['cursor'] = cursor
        params = flatten_params({
            'variables': variables,
            'features': features
        })
        return await self.base.get(
            Endpoint.BOOKMARKS,
            params=params,
            headers=self.base._base_headers
        )

    async def bookmark_folder_timeline(self, count, cursor, folder_id):
        variables = {
            'count': count,
            'includePromotedContent': True,
            'bookmark_collection_id': folder_id
        }
        variables['bookmark_collection_id'] = folder_id
        if cursor is not None:
            variables['cursor'] = cursor
        return await self.gql_get(Endpoint.BOOKMARK_FOLDER_TIMELINE, variables, BOOKMARK_FOLDER_TIMELINE_FEATURES)

    async def delete_all_bookmarks(self):
        return await self.gql_post(Endpoint.BOOKMARKS_ALL_DELETE, {})

    async def bookmark_folders_slice(self, cursor):
        variables = {}
        if cursor is not None:
            variables['cursor'] = cursor
        variables = {'variables': variables}
        return await self.gql_get(Endpoint.BOOKMARK_FOLDERS_SLICE, variables)

    async def edit_bookmark_folder(self, folder_id, name):
        variables = {
            'bookmark_collection_id': folder_id,
            'name': name
        }
        return await self.gql_post(Endpoint.EDIT_BOOKMARK_FOLDER, variables)

    async def delete_bookmark_folder(self, folder_id):
        variables = {'bookmark_collection_id': folder_id}
        return await self.gql_post(Endpoint.DELETE_BOOKMARK_FOLDER, variables)

    async def create_bookmark_folder(self, name):
        variables = {'name': name}
        return await self.gql_post(Endpoint.CREATE_BOOKMARK_FOLDER, variables)

    async def _friendships(self, user_id, count, endpoint, cursor):
        variables = {
            'userId': user_id,
            'count': count,
            'includePromotedContent': False
        }
        if cursor is not None:
            variables['cursor'] = cursor
        return await self.gql_get(endpoint, variables, FEATURES)

    async def followers(self, user_id, count, cursor):
        return await self._friendships(user_id, count, Endpoint.FOLLOWERS, cursor)

    async def blue_verified_followers(self, user_id, count, cursor):
        return await self._friendships(user_id, count, Endpoint.BLUE_VERIFIED_FOLLOWERS, cursor)

    async def followers_you_know(self, user_id, count, cursor):
        return await self._friendships(user_id, count, Endpoint.FOLLOWERS_YOU_KNOW, cursor)

    async def following(self, user_id, count, cursor):
        return await self._friendships(user_id, count, Endpoint.FOLLOWING, cursor)

    async def user_creator_subscriptions(self, user_id, count, cursor):
        return await self._friendships(user_id, count, Endpoint.USER_CREATOR_SUBSCRIPTIONS, cursor)

    async def user_dm_reaction_mutation_add_mutation(self, message_id, conversation_id, emoji):
        variables = {
            'messageId': message_id,
            'conversationId': conversation_id,
            'reactionTypes': ['Emoji'],
            'emojiReactions': [emoji]
        }
        return await self.gql_post(Endpoint.USER_DM_REACTION_MUTATION_ADD_MUTATION, variables)

    async def user_dm_reaction_mutation_remove_mutation(self, message_id, conversation_id, emoji):
        variables = {
            'conversationId': conversation_id,
            'messageId': message_id,
            'reactionTypes': ['Emoji'],
            'emojiReactions': [emoji]
        }
        return await self.gql_post(Endpoint.USER_DM_REACTION_MUTATION_REMOVE_MUTATION, variables)

    async def dm_message_delete_mutation(self, message_id):
        variables = {'messageId': message_id}
        return await self.gql_post(Endpoint.DM_MESSAGE_DELETE_MUTATION, variables)

    async def add_participants_mutation(self, group_id, user_ids):
        variables = {
            'addedParticipants': user_ids,
            'conversationId': group_id
        }
        return await self.gql_post(Endpoint.ADD_PARTICIPANTS_MUTATION, variables)

    async def create_list(self, name, description, is_private):
        variables = {
            'isPrivate': is_private,
            'name': name,
            'description': description
        }
        return await self.gql_post(Endpoint.CREATE_LIST, variables, LIST_FEATURES)

    async def edit_list_banner(self, list_id, media_id):
        variables = {
            'listId': list_id,
            'mediaId': media_id
        }
        return await self.gql_post(Endpoint.EDIT_LIST_BANNER, variables, LIST_FEATURES)

    async def delete_list_banner(self, list_id):
        variables = {'listId': list_id}
        return await self.gql_post(Endpoint.DELETE_LIST_BANNER, variables, LIST_FEATURES)

    async def update_list(self, list_id, name, description, is_private):
        variables = {'listId': list_id}
        if name is not None:
            variables['name'] = name
        if description is not None:
            variables['description'] = description
        if is_private is not None:
            variables['isPrivate'] = is_private
        return await self.gql_post(Endpoint.UPDATE_LIST, variables, LIST_FEATURES)

    async def list_add_member(self, list_id, user_id):
        variables = {
            'listId': list_id,
            'userId': user_id
        }
        return await self.gql_post(Endpoint.LIST_ADD_MEMBER, variables, LIST_FEATURES)

    async def list_remove_member(self, list_id, user_id):
        variables = {
            'listId': list_id,
            'userId': user_id
        }
        return await self.gql_post(Endpoint.LIST_REMOVE_MEMBER, variables, LIST_FEATURES)

    async def list_management_pace_timeline(self, count, cursor):
        variables = {'count': count}
        if cursor is not None:
            variables['cursor'] = cursor
        return await self.gql_get(Endpoint.LIST_MANAGEMENT_PACE_TIMELINE, variables, FEATURES)

    async def list_by_rest_id(self, list_id):
        variables = {'listId': list_id}
        return await self.gql_get(Endpoint.LIST_BY_REST_ID, variables, LIST_FEATURES)

    async def list_latest_tweets_timeline(self, list_id, count, cursor):
        variables = {'listId': list_id, 'count': count}
        if cursor is not None:
            variables['cursor'] = cursor
        return await self.gql_get(Endpoint.LIST_LATEST_TWEETS_TIMELINE, variables, FEATURES)

    async def _list_users(self, endpoint, list_id, count, cursor):
        variables = {'listId': list_id, 'count': count}
        if cursor is not None:
            variables['cursor'] = cursor
        return await self.gql_get(endpoint, variables, FEATURES)

    async def list_members(self, list_id, count, cursor):
        return await self._list_users(Endpoint.LIST_MEMBERS, list_id, count, cursor)

    async def list_subscribers(self, list_id, count, cursor):
        return await self._list_users(Endpoint.LIST_SUBSCRIBERS, list_id, count, cursor)

    async def search_community(self, query, cursor):
        variables = {'query': query}
        if cursor is not None:
            variables['cursor'] = cursor
        return await self.gql_get(Endpoint.SEARCH_COMMUNITY, variables)

    async def community_query(self, community_id):
        variables = {'communityId': community_id}
        features = {
            'c9s_list_members_action_api_enabled': False,
            'c9s_superc9s_indication_enabled': False
        }
        return await self.gql_get(Endpoint.COMMUNITY_QUERY, variables, features)

    async def community_media_timeline(self, community_id, count, cursor):
        variables = {
            'communityId': community_id,
            'count': count,
            'withCommunity': True
        }
        if cursor is not None:
            variables['cursor'] = cursor
        return await self.gql_get(Endpoint.COMMUNITY_MEDIA_TIMELINE, variables, COMMUNITY_TWEETS_FEATURES)

    async def community_tweets_timeline(self, community_id, ranking_mode, count, cursor):
        variables = {
            'communityId': community_id,
            'count': count,
            'withCommunity': True,
            'rankingMode': ranking_mode
        }
        if cursor is not None:
            variables['cursor'] = cursor
        return await self.gql_get(Endpoint.COMMUNITY_TWEETS_TIMELINE, variables, COMMUNITY_TWEETS_FEATURES)

    async def communities_main_page_timeline(self, count, cursor):
        variables = {
            'count': count,
            'withCommunity': True
        }
        if cursor is not None:
            variables['cursor'] = cursor
        return await self.gql_get(Endpoint.COMMUNITIES_MAIN_PAGE_TIMELINE, variables, COMMUNITY_TWEETS_FEATURES)

    async def join_community(self, community_id):
        variables = {'communityId': community_id}
        return await self.gql_post(Endpoint.JOIN_COMMUNITY, variables, JOIN_COMMUNITY_FEATURES)

    async def leave_community(self, community_id):
        variables = {'communityId': community_id}
        return await self.gql_post(Endpoint.LEAVE_COMMUNITY, variables, JOIN_COMMUNITY_FEATURES)

    async def request_to_join_community(self, community_id, answer):
        variables = {
            'communityId': community_id,
            'answer': '' if answer is None else answer
        }
        return await self.gql_post(Endpoint.REQUEST_TO_JOIN_COMMUNITY, variables, JOIN_COMMUNITY_FEATURES)

    async def _get_community_users(self, endpoint, community_id, count, cursor):
        variables = {'communityId': community_id, 'count': count}
        features = {'responsive_web_graphql_timeline_navigation_enabled': True}
        if cursor is not None:
            variables['cursor'] = cursor
        return await self.gql_get(endpoint, variables, features)

    async def members_slice_timeline_query(self, community_id, count, cursor):
        return await self._get_community_users(Endpoint.MEMBERS_SLICE_TIMELINE_QUERY, community_id, count, cursor)

    async def moderators_slice_timeline_query(self, community_id, count, cursor):
        return await self._get_community_users(Endpoint.MODERATORS_SLICE_TIMELINE_QUERY, community_id, count, cursor)

    async def community_tweet_search_module_query(self, community_id, query, count, cursor):
        variables = {
            'count': count,
            'query': query,
            'communityId': community_id,
            'includePromotedContent': False,
            'withBirdwatchNotes': True,
            'withVoice': False,
            'isListMemberTargetUserId': '0',
            'withCommunity': False,
            'withSafetyModeUserFields': True
        }
        if cursor is not None:
            variables['cursor'] = cursor
        return await self.gql_get(Endpoint.COMMUNITY_TWEET_SEARCH_MODULE_QUERY, variables, COMMUNITY_TWEETS_FEATURES)

    async def tweet_results_by_rest_ids(self, tweet_ids):
        variables = {
            'tweetIds': tweet_ids,
            'includePromotedContent': True,
            'withBirdwatchNotes': True,
            'withVoice': True,
            'withCommunity': True
        }
        return await self.gql_get(Endpoint.TWEET_RESULTS_BY_REST_IDS, variables, TWEET_RESULTS_BY_REST_IDS_FEATURES)

    ####################
    # For guest client
    ####################

    async def tweet_result_by_rest_id(self, tweet_id):
        variables = {
            'tweetId': tweet_id,
            'withCommunity': False,
            'includePromotedContent': False,
            'withVoice': False
        }
        params = {
            'fieldToggles': {
                'withArticleRichContentState': True,
                'withArticlePlainText': False,
                'withGrokAnalyze': False
            }
        }
        return await self.gql_get(
            Endpoint.TWEET_RESULT_BY_REST_ID, variables, TWEET_RESULT_BY_REST_ID_FEATURES, extra_params=params
        )



from __future__ import annotations

from typing import TYPE_CHECKING

from ..media import MEDIA_TYPE, _media_from_data
from ..utils import find_dict
from .user import User

if TYPE_CHECKING:
    from .client import GuestClient


class Tweet:
    """
    Attributes
    ----------
    id : :class:`str`
        The unique identifier of the tweet.
    created_at : :class:`str`
        The date and time when the tweet was created.
    created_at_datetime : :class:`datetime`
        The created_at converted to datetime.
    user: :class:`.guest.user.User`
        Author of the tweet.
    text : :class:`str`
        The full text of the tweet.
    lang : :class:`str`
        The language of the tweet.
    in_reply_to : :class:`str`
        The tweet ID this tweet is in reply to, if any
    is_quote_status : :class:`bool`
        Indicates if the tweet is a quote status.
    quote : :class:`.guest.tweet.Tweet` | None
        The Tweet being quoted (if any)
    retweeted_tweet : :class:`.guest.tweet.Tweet` | None
        The Tweet being retweeted (if any)
    possibly_sensitive : :class:`bool`
        Indicates if the tweet content may be sensitive.
    possibly_sensitive_editable : :class:`bool`
        Indicates if the tweet's sensitivity can be edited.
    quote_count : :class:`int`
        The count of quotes for the tweet.
    media : list[:class:`.media.Photo` | :class:`.media.AnimatedGif` | :class:`.media.Video`]
        A list of media entities associated with the tweet.
    reply_count : :class:`int`
        The count of replies to the tweet.
    favorite_count : :class:`int`
        The count of favorites or likes for the tweet.
    favorited : :class:`bool`
        Indicates if the tweet is favorited.
    view_count: :class:`int` | None
        The count of views.
    view_count_state : :class:`str` | None
        The state of the tweet views.
    retweet_count : :class:`int`
        The count of retweets for the tweet.
    bookmark_count : :class:`int`
        The count of bookmarks for the tweet.
    bookmarked : :class:`bool`
        Indicates if the tweet is bookmarked.
    place : :class:`.Place` | None
        The location associated with the tweet.
    editable_until_msecs : :class:`int`
        The timestamp until which the tweet is editable.
    is_translatable : :class:`bool`
        Indicates if the tweet is translatable.
    is_edit_eligible : :class:`bool`
        Indicates if the tweet is eligible for editing.
    edits_remaining : :class:`int`
        The remaining number of edits allowed for the tweet.
    edit_tweet_ids : :class:`list`[:class:`int`]
        List of tweet IDs representing the edit history of the tweet.
    reply_to: list[:class:`Tweet`] | None
        A list of Tweet objects representing the tweets to which to reply.
    related_tweets : list[:class:`Tweet`] | None
        Related tweets.
    hashtags: list[:class:`str`]
        Hashtags included in the tweet text.
    has_card : :class:`bool`
        Indicates if the tweet contains a card.
    thumbnail_title : :class:`str` | None
        The title of the webpage displayed inside tweet's card.
    thumbnail_url : :class:`str` | None
        Link to the image displayed in the tweet's card.
    urls : :class:`list`
        Information about URLs contained in the tweet.
    full_text : :class:`str` | None
        The full text of the tweet.
    """

    def __init__(self, client: GuestClient, data: dict, user: User = None) -> None:
        self._client = client
        self._data = data
        self.user = user

        self.reply_to: list[Tweet] | None = None
        self.related_tweets: list[Tweet] | None = None
        self.thread: list[Tweet] | None = None

        self.id: str = data['rest_id']
        legacy = data['legacy']
        self.created_at: str = legacy['created_at']
        self.text: str = legacy['full_text']
        self.lang: str = legacy['lang']
        self.is_quote_status: bool = legacy['is_quote_status']
        self.in_reply_to: str | None = self._data['legacy'].get('in_reply_to_status_id_str')
        self.is_quote_status: bool = legacy['is_quote_status']
        self.possibly_sensitive: bool = legacy.get('possibly_sensitive')
        self.possibly_sensitive_editable: bool = legacy.get('possibly_sensitive_editable')
        self.quote_count: int = legacy['quote_count']
        self._media: list = legacy['entities'].get('media')
        self.reply_count: int = legacy['reply_count']
        self.favorite_count: int = legacy['favorite_count']
        self.favorited: bool = legacy['favorited']
        self.retweet_count: int = legacy['retweet_count']
        self._place_data = legacy.get('place')
        self.bookmark_count: int = legacy.get('bookmark_count')
        self.bookmarked: bool = legacy.get('bookmarked')
        self.edit_tweet_ids: list[int] = data['edit_control'].get('edit_tweet_ids', [])
        self.editable_until_msecs: int = data['edit_control'].get('editable_until_msecs')
        self.is_translatable: bool = data.get('is_translatable')
        self.is_edit_eligible: bool = data['edit_control'].get('is_edit_eligible')
        self.edits_remaining: int = data['edit_control'].get('edits_remaining')
        self.view_count: str = data['views'].get('count') if 'views' in data else None
        self.view_count_state: str = data['views'].get('state') if 'views' in data else None
        self.has_community_notes: bool = data.get('has_birdwatch_notes')

        if data.get('quoted_status_result'):
            quoted_tweet = data.pop('quoted_status_result')['result']
            if 'tweet' in quoted_tweet:
                quoted_tweet = quoted_tweet['tweet']
            if quoted_tweet.get('__typename') != 'TweetTombstone':
                quoted_user = User(client, quoted_tweet['core']['user_results']['result'])
                self.quote: Tweet = Tweet(client, quoted_tweet, quoted_user)
        else:
            self.quote = None

        if legacy.get('retweeted_status_result'):
            retweeted_tweet = legacy.pop('retweeted_status_result')['result']
            if 'tweet' in retweeted_tweet:
                retweeted_tweet = retweeted_tweet['tweet']
            retweeted_user = User(
                client, retweeted_tweet['core']['user_results']['result']
            )
            self.retweeted_tweet: Tweet = Tweet(
                client, retweeted_tweet, retweeted_user
            )
        else:
            self.retweeted_tweet = None

        note_tweet_results = find_dict(data, 'note_tweet_results', find_one=True)
        self.full_text: str = self.text
        if note_tweet_results:
            text_list = find_dict(note_tweet_results, 'text', find_one=True)
            if text_list:
                self.full_text = text_list[0]

            entity_set = note_tweet_results[0]['result']['entity_set']
            self.urls: list = entity_set.get('urls')
            hashtags = entity_set.get('hashtags', [])
        else:
            self.urls: list = legacy['entities'].get('urls')
            hashtags = legacy['entities'].get('hashtags', [])

        self.hashtags: list[str] = [
            i['text'] for i in hashtags
        ]

        self.community_note = None
        if 'birdwatch_pivot' in data:
            community_note_data = data['birdwatch_pivot']
            if 'note' in community_note_data:
                self.community_note = {
                    'id': community_note_data['note']['rest_id'],
                    'text': community_note_data['subtitle']['text']
                }

        if (
            'card' in data and
            'legacy' in data['card'] and
            'name' in data['card']['legacy'] and
            data['card']['legacy']['name'].startswith('poll')
        ):
            self._poll_data = data['card']
        else:
            self._poll_data = None

        self.thumbnail_url = None
        self.thumbnail_title = None
        self.has_card = 'card' in data
        if (
            'card' in data and
            'legacy' in data['card'] and
            'binding_values' in data['card']['legacy']
        ):
            card_data = data['card']['legacy']['binding_values']

            if isinstance(card_data, list):
                binding_values = {
                    i.get('key'): i.get('value')
                    for i in card_data
                }

            if 'title' in binding_values and 'string_value' in binding_values['title']:
                self.thumbnail_title = binding_values['title']['string_value']

            if (
                'thumbnail_image_original' in binding_values and
                'image_value' in binding_values['thumbnail_image_original'] and
                'url' in binding_values['thumbnail_image_original']['image_value']
            ):
                self.thumbnail_url = binding_values['thumbnail_image_original']['image_value']['url']

    @property
    def media(self) -> list[MEDIA_TYPE]:
        m = []
        for entry in self._media:
            media_obj = _media_from_data(self._client, entry)
            if not media_obj:
                continue
            m.append(media_obj)
        return m

    async def update(self) -> None:
        new = await self._client.get_tweet_by_id(self.id)
        self.__dict__.update(new.__dict__)

    def __repr__(self) -> str:
        return f'<Tweet id="{self.id}">'

    def __eq__(self, __value: object) -> bool:
        return isinstance(__value, Tweet) and self.id == __value.id

    def __ne__(self, __value: object) -> bool:
        return not self == __value


from __future__ import annotations

import re
from typing import TYPE_CHECKING, NamedTuple

from bs4 import BeautifulSoup
from httpx import Response
from ..constants import DOMAIN

if TYPE_CHECKING:
    from ..client.client import Client


class UnlockHTML(NamedTuple):
    authenticity_token: str
    assignment_token: str
    needs_unlock: bool
    start_button: bool
    finish_button: bool
    delete_button: bool
    blob: str


class CaptchaSolver:
    client: Client
    max_attempts: int

    CAPTCHA_URL = f'https://{DOMAIN}/account/access'
    CAPTCHA_SITE_KEY = '0152B4EB-D2DC-460A-89A1-629838B529C9'

    async def get_unlock_html(self) -> tuple[Response, UnlockHTML]:
        headers = {
            'X-Twitter-Client-Language': 'en-US',
            'User-Agent': self.client._user_agent,
            'Upgrade-Insecure-Requests': '1'
        }
        _, response = await self.client.get(
            self.CAPTCHA_URL, headers=headers
        )
        return response, parse_unlock_html(response.text)

    async def ui_metrix(self) -> str:
        js, _ = await self.client.get(
            f'https://{DOMAIN}/i/js_inst?c_name=ui_metrics'
        )
        return re.findall(r'return ({.*?});', js, re.DOTALL)[0]

    async def confirm_unlock(
        self,
        authenticity_token: str,
        assignment_token: str,
        verification_string: str = None,
        ui_metrics: bool = False
    ) -> tuple[Response, UnlockHTML]:
        data = {
            'authenticity_token': authenticity_token,
            'assignment_token': assignment_token,
            'lang': 'en',
            'flow': '',
        }
        params = {}
        if verification_string:
            data['verification_string'] = verification_string
            data['language_code'] = 'en'
            params['lang'] = 'en'
        if ui_metrics:
            data['ui_metrics'] = await self.client._ui_metrix()
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded',
            'Upgrade-Insecure-Requests': '1',
            'Referer': self.CAPTCHA_URL
        }
        _, response = await self.client.post(
            self.CAPTCHA_URL, params=params, data=data, headers=headers
        )
        return response, parse_unlock_html(response.text)


def parse_unlock_html(html: str) -> UnlockHTML:
    soup = BeautifulSoup(html, 'lxml')

    authenticity_token = None
    authenticity_token_element = soup.find(
        'input', {'name': 'authenticity_token'}
    )
    if authenticity_token_element is not None:
        authenticity_token: str = authenticity_token_element.get('value')

    assignment_token = None
    assignment_token_element = soup.find('input', {'name': 'assignment_token'})
    if assignment_token_element is not None:
        assignment_token = assignment_token_element.get('value')

    verification_string = soup.find('input', id='verification_string')
    needs_unlock = bool(verification_string)
    start_button = bool(soup.find('input', value='Start'))
    finish_button = bool(soup.find('input', value='Continue to X'))
    delete_button = bool(soup.find('input', value='Delete'))

    iframe = soup.find(id='arkose_iframe')
    blob = re.findall(r'data=(.+)', iframe['src'])[0] if iframe else None

    return UnlockHTML(
        authenticity_token,
        assignment_token,
        needs_unlock,
        start_button,
        finish_button,
        delete_button,
        blob
    )



from typing import Union, List


class Cubic:
    def __init__(self, curves: List[Union[float, int]]):
        self.curves = curves

    def get_value(self, time: Union[float, int]):
        start_gradient = end_gradient = start = mid = 0.0
        end = 1.0

        if time <= 0.0:
            if self.curves[0] > 0.0:
                start_gradient = self.curves[1] / self.curves[0]
            elif self.curves[1] == 0.0 and self.curves[2] > 0.0:
                start_gradient = self.curves[3] / self.curves[2]
            return start_gradient * time

        if time >= 1.0:
            if self.curves[2] < 1.0:
                end_gradient = (self.curves[3] - 1.0) / (self.curves[2] - 1.0)
            elif self.curves[2] == 1.0 and self.curves[0] < 1.0:
                end_gradient = (self.curves[1] - 1.0) / (self.curves[0] - 1.0)
            return 1.0 + end_gradient * (time - 1.0)

        while start < end:
            mid = (start + end) / 2
            x_est = self.calculate(self.curves[0], self.curves[2], mid)
            if abs(time - x_est) < 0.00001:
                return self.calculate(self.curves[1], self.curves[3], mid)
            if x_est < time:
                start = mid
            else:
                end = mid
        return self.calculate(self.curves[1], self.curves[3], mid)

    @staticmethod
    def calculate(a, b, m):
        return 3.0 * a * (1 - m) * (1 - m) * m + 3.0 * b * (1 - m) * m * m + m * m * m

# Example usage:
# cubic_instance = Cubic([0.1, 0.2, 0.3, 0.4])
# value = cubic_instance.get_value(0.5)
# print(value)


if __name__ == "__main__":
    pass



# This token is common to all accounts and does not need to be changed.
TOKEN = 'AAAAAAAAAAAAAAAAAAAAANRILgAAAAAAnNwIzUejRCOuH5E6I8xnZz4puTs%3D1Zv7ttfk8LF81IUq16cHjhLTvJu4FA33AGWWjCpTnA'

DOMAIN = 'x.com'

FEATURES = {
    'creator_subscriptions_tweet_preview_api_enabled': True,
    'c9s_tweet_anatomy_moderator_badge_enabled': True,
    'tweetypie_unmention_optimization_enabled': True,
    'responsive_web_edit_tweet_api_enabled': True,
    'graphql_is_translatable_rweb_tweet_is_translatable_enabled': True,
    'view_counts_everywhere_api_enabled': True,
    'longform_notetweets_consumption_enabled': True,
    'responsive_web_twitter_article_tweet_consumption_enabled': True,
    'tweet_awards_web_tipping_enabled': False,
    'longform_notetweets_rich_text_read_enabled': True,
    'longform_notetweets_inline_media_enabled': True,
    'rweb_video_timestamps_enabled': True,
    'responsive_web_graphql_exclude_directive_enabled': True,
    'verified_phone_label_enabled': False,
    'freedom_of_speech_not_reach_fetch_enabled': True,
    'standardized_nudges_misinfo': True,
    'tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled': True,
    'responsive_web_media_download_video_enabled': False,
    'responsive_web_graphql_skip_user_profile_image_extensions_enabled': False,
    'responsive_web_graphql_timeline_navigation_enabled': True,
    'responsive_web_enhance_cards_enabled': False
}

USER_FEATURES = {
    'hidden_profile_likes_enabled': True,
    'hidden_profile_subscriptions_enabled': True,
    'responsive_web_graphql_exclude_directive_enabled': True,
    'verified_phone_label_enabled': False,
    'subscriptions_verification_info_is_identity_verified_enabled': True,
    'subscriptions_verification_info_verified_since_enabled': True,
    'highlights_tweets_tab_ui_enabled': True,
    'responsive_web_twitter_article_notes_tab_enabled': False,
    'creator_subscriptions_tweet_preview_api_enabled': True,
    'responsive_web_graphql_skip_user_profile_image_extensions_enabled': False,
    'responsive_web_graphql_timeline_navigation_enabled': True
}

LIST_FEATURES = {
    'responsive_web_graphql_exclude_directive_enabled': True,
    'verified_phone_label_enabled': False,
    'responsive_web_graphql_skip_user_profile_image_extensions_enabled': False,
    'responsive_web_graphql_timeline_navigation_enabled': True
}

COMMUNITY_NOTE_FEATURES = {
    'responsive_web_birdwatch_media_notes_enabled': True,
    'responsive_web_graphql_timeline_navigation_enabled': True,
    'rweb_tipjar_consumption_enabled': False,
    'responsive_web_graphql_exclude_directive_enabled': True,
    'verified_phone_label_enabled': False,
    'responsive_web_graphql_skip_user_profile_image_extensions_enabled': False
}

COMMUNITY_TWEETS_FEATURES = {
    'rweb_tipjar_consumption_enabled': True,
    'responsive_web_graphql_exclude_directive_enabled': True,
    'verified_phone_label_enabled': False,
    'creator_subscriptions_tweet_preview_api_enabled': True,
    'responsive_web_graphql_timeline_navigation_enabled': True,
    'responsive_web_graphql_skip_user_profile_image_extensions_enabled': False,
    'communities_web_enable_tweet_community_results_fetch': True,
    'c9s_tweet_anatomy_moderator_badge_enabled': True,
    'tweetypie_unmention_optimization_enabled': True,
    'responsive_web_edit_tweet_api_enabled': True,
    'graphql_is_translatable_rweb_tweet_is_translatable_enabled': True,
    'view_counts_everywhere_api_enabled': True,
    'longform_notetweets_consumption_enabled': True,
    'responsive_web_twitter_article_tweet_consumption_enabled': True,
    'tweet_awards_web_tipping_enabled': False,
    'creator_subscriptions_quote_tweet_preview_enabled': False,
    'freedom_of_speech_not_reach_fetch_enabled': True,
    'standardized_nudges_misinfo': True,
    'tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled': True,
    'rweb_video_timestamps_enabled': True,
    'longform_notetweets_rich_text_read_enabled': True,
    'longform_notetweets_inline_media_enabled': True,
    'responsive_web_enhance_cards_enabled': False
}

JOIN_COMMUNITY_FEATURES = {
    'rweb_tipjar_consumption_enabled': True,
    'responsive_web_graphql_exclude_directive_enabled': True,
    'verified_phone_label_enabled': False,
    'responsive_web_graphql_skip_user_profile_image_extensions_enabled': False,
    'responsive_web_graphql_timeline_navigation_enabled': True
}

NOTE_TWEET_FEATURES = {
    'communities_web_enable_tweet_community_results_fetch': True,
    'c9s_tweet_anatomy_moderator_badge_enabled': True,
    'tweetypie_unmention_optimization_enabled': True,
    'responsive_web_edit_tweet_api_enabled': True,
    'graphql_is_translatable_rweb_tweet_is_translatable_enabled': True,
    'view_counts_everywhere_api_enabled': True,
    'longform_notetweets_consumption_enabled': True,
    'responsive_web_twitter_article_tweet_consumption_enabled': True,
    'tweet_awards_web_tipping_enabled': False,
    'creator_subscriptions_quote_tweet_preview_enabled': False,
    'longform_notetweets_rich_text_read_enabled': True,
    'longform_notetweets_inline_media_enabled': True,
    'articles_preview_enabled': False,
    'rweb_video_timestamps_enabled': True,
    'rweb_tipjar_consumption_enabled': True,
    'responsive_web_graphql_exclude_directive_enabled': True,
    'verified_phone_label_enabled': False,
    'freedom_of_speech_not_reach_fetch_enabled': True,
    'standardized_nudges_misinfo': True,
    'tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled': True,
    'tweet_with_visibility_results_prefer_gql_media_interstitial_enabled': True,
    'responsive_web_graphql_skip_user_profile_image_extensions_enabled': False,
    'responsive_web_graphql_timeline_navigation_enabled': True,
    'responsive_web_enhance_cards_enabled': False
}

SIMILAR_POSTS_FEATURES = {
    'rweb_tipjar_consumption_enabled': True,
    'responsive_web_graphql_exclude_directive_enabled': True,
    'verified_phone_label_enabled': False,
    'creator_subscriptions_tweet_preview_api_enabled': True,
    'responsive_web_graphql_timeline_navigation_enabled': True,
    'responsive_web_graphql_skip_user_profile_image_extensions_enabled': False,
    'communities_web_enable_tweet_community_results_fetch': True,
    'c9s_tweet_anatomy_moderator_badge_enabled': True,
    'articles_preview_enabled': False,
    'tweetypie_unmention_optimization_enabled': True,
    'responsive_web_edit_tweet_api_enabled': True,
    'graphql_is_translatable_rweb_tweet_is_translatable_enabled': True,
    'view_counts_everywhere_api_enabled': True,
    'longform_notetweets_consumption_enabled': True,
    'responsive_web_twitter_article_tweet_consumption_enabled': True,
    'tweet_awards_web_tipping_enabled': False,
    'creator_subscriptions_quote_tweet_preview_enabled': False,
    'freedom_of_speech_not_reach_fetch_enabled': True,
    'standardized_nudges_misinfo': True,
    'tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled': True,
    'tweet_with_visibility_results_prefer_gql_media_interstitial_enabled': True,
    'rweb_video_timestamps_enabled': True,
    'longform_notetweets_rich_text_read_enabled': True,
    'longform_notetweets_inline_media_enabled': True,
    'responsive_web_enhance_cards_enabled': False
}

BOOKMARK_FOLDER_TIMELINE_FEATURES = {
    'rweb_tipjar_consumption_enabled': True,
    'responsive_web_graphql_exclude_directive_enabled': True,
    'verified_phone_label_enabled': False,
    'creator_subscriptions_tweet_preview_api_enabled': True,
    'responsive_web_graphql_timeline_navigation_enabled': True,
    'responsive_web_graphql_skip_user_profile_image_extensions_enabled': False,
    'communities_web_enable_tweet_community_results_fetch': True,
    'c9s_tweet_anatomy_moderator_badge_enabled': True,
    'articles_preview_enabled': False,
    'tweetypie_unmention_optimization_enabled': True,
    'responsive_web_edit_tweet_api_enabled': True,
    'graphql_is_translatable_rweb_tweet_is_translatable_enabled': True,
    'view_counts_everywhere_api_enabled': True,
    'longform_notetweets_consumption_enabled': True,
    'responsive_web_twitter_article_tweet_consumption_enabled': True,
    'tweet_awards_web_tipping_enabled': False,
    'creator_subscriptions_quote_tweet_preview_enabled': False,
    'freedom_of_speech_not_reach_fetch_enabled': True,
    'standardized_nudges_misinfo': True,
    'tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled': True,
    'tweet_with_visibility_results_prefer_gql_media_interstitial_enabled': True,
    'rweb_video_timestamps_enabled': True,
    'longform_notetweets_rich_text_read_enabled': True,
    'longform_notetweets_inline_media_enabled': True,
    'responsive_web_enhance_cards_enabled': False
}

TWEET_RESULT_BY_REST_ID_FEATURES = {
    'creator_subscriptions_tweet_preview_api_enabled': True,
    'communities_web_enable_tweet_community_results_fetch': True,
    'c9s_tweet_anatomy_moderator_badge_enabled': True,
    'articles_preview_enabled': True,
    'tweetypie_unmention_optimization_enabled': True,
    'responsive_web_edit_tweet_api_enabled': True,
    'graphql_is_translatable_rweb_tweet_is_translatable_enabled': True,
    'view_counts_everywhere_api_enabled': True,
    'longform_notetweets_consumption_enabled': True,
    'responsive_web_twitter_article_tweet_consumption_enabled': True,
    'tweet_awards_web_tipping_enabled': False,
    'creator_subscriptions_quote_tweet_preview_enabled': False,
    'freedom_of_speech_not_reach_fetch_enabled': True,
    'standardized_nudges_misinfo': True,
    'tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled': True,
    'rweb_video_timestamps_enabled': True,
    'longform_notetweets_rich_text_read_enabled': True,
    'longform_notetweets_inline_media_enabled': True,
    'rweb_tipjar_consumption_enabled': True,
    'responsive_web_graphql_exclude_directive_enabled': True,
    'verified_phone_label_enabled': False,
    'responsive_web_graphql_skip_user_profile_image_extensions_enabled': False,
    'responsive_web_graphql_timeline_navigation_enabled': True,
    'responsive_web_enhance_cards_enabled': False
}

USER_HIGHLIGHTS_TWEETS_FEATURES = {
    'rweb_tipjar_consumption_enabled': True,
    'responsive_web_graphql_exclude_directive_enabled': True,
    'verified_phone_label_enabled': False,
    'creator_subscriptions_tweet_preview_api_enabled': True,
    'responsive_web_graphql_timeline_navigation_enabled': True,
    'responsive_web_graphql_skip_user_profile_image_extensions_enabled': False,
    'communities_web_enable_tweet_community_results_fetch': True,
    'c9s_tweet_anatomy_moderator_badge_enabled': True,
    'articles_preview_enabled': True,
    'tweetypie_unmention_optimization_enabled': True,
    'responsive_web_edit_tweet_api_enabled': True,
    'graphql_is_translatable_rweb_tweet_is_translatable_enabled': True,
    'view_counts_everywhere_api_enabled': True,
    'longform_notetweets_consumption_enabled': True,
    'responsive_web_twitter_article_tweet_consumption_enabled': True,
    'tweet_awards_web_tipping_enabled': False,
    'creator_subscriptions_quote_tweet_preview_enabled': False,
    'freedom_of_speech_not_reach_fetch_enabled': True,
    'standardized_nudges_misinfo': True,
    'tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled': True,
    'rweb_video_timestamps_enabled': True,
    'longform_notetweets_rich_text_read_enabled': True,
    'longform_notetweets_inline_media_enabled': True,
    'responsive_web_enhance_cards_enabled': False
}

TWEET_RESULTS_BY_REST_IDS_FEATURES = {
    'creator_subscriptions_tweet_preview_api_enabled': True,
    'premium_content_api_read_enabled': False,
    'communities_web_enable_tweet_community_results_fetch': True,
    'c9s_tweet_anatomy_moderator_badge_enabled': True,
    'responsive_web_grok_analyze_button_fetch_trends_enabled': False,
    'responsive_web_grok_analyze_post_followups_enabled': True,
    'responsive_web_grok_share_attachment_enabled': True,
    'articles_preview_enabled': True,
    'responsive_web_edit_tweet_api_enabled': True,
    'graphql_is_translatable_rweb_tweet_is_translatable_enabled': True,
    'view_counts_everywhere_api_enabled': True,
    'longform_notetweets_consumption_enabled': True,
    'responsive_web_twitter_article_tweet_consumption_enabled': True,
    'tweet_awards_web_tipping_enabled': False,
    'creator_subscriptions_quote_tweet_preview_enabled': False,
    'freedom_of_speech_not_reach_fetch_enabled': True,
    'standardized_nudges_misinfo': True,
    'tweet_with_visibility_results_prefer_gql_limited_actions_policy_enabled': True,
    'rweb_video_timestamps_enabled': True,
    'longform_notetweets_rich_text_read_enabled': True,
    'longform_notetweets_inline_media_enabled': True,
    'profile_label_improvements_pcf_label_in_post_enabled': False,
    'rweb_tipjar_consumption_enabled': True,
    'responsive_web_graphql_exclude_directive_enabled': True,
    'verified_phone_label_enabled': False,
    'responsive_web_graphql_skip_user_profile_image_extensions_enabled': False,
    'responsive_web_graphql_timeline_navigation_enabled': True,
    'responsive_web_enhance_cards_enabled': False
}



from __future__ import annotations

from typing import TYPE_CHECKING, Literal, NamedTuple

from .tweet import Tweet
from .user import User
from .utils import Result, b64_to_str

if TYPE_CHECKING:
    from .client.client import Client


class CommunityCreator(NamedTuple):
    id: str
    screen_name: str
    verified: bool


class CommunityRule(NamedTuple):
    id: str
    name: str


class CommunityMember:
    def __init__(self, client: Client, data: dict) -> None:
        self._client = client
        self.id: str = data['rest_id']

        self.community_role: str = data['community_role']
        self.super_following: bool = data['super_following']
        self.super_follow_eligible: bool = data['super_follow_eligible']
        self.super_followed_by: bool = data['super_followed_by']
        self.smart_blocking: bool = data['smart_blocking']
        self.is_blue_verified: bool = data['is_blue_verified']

        legacy = data['legacy']
        self.screen_name: str = legacy['screen_name']
        self.name: str = legacy['name']
        self.follow_request_sent: bool = legacy['follow_request_sent']
        self.protected: bool = legacy['protected']
        self.following: bool = legacy['following']
        self.followed_by: bool = legacy['followed_by']
        self.blocking: bool = legacy['blocking']
        self.profile_image_url_https: str = legacy['profile_image_url_https']
        self.verified: bool = legacy['verified']

    def __eq__(self, __value: object) -> bool:
        return isinstance(__value, CommunityMember) and self.id == __value.id

    def __ne__(self, __value: object) -> bool:
        return not self == __value

    def __repr__(self) -> str:
        return f'<CommunityMember id="{self.id}">'


class Community:
    """
    Attributes
    ----------
    id : :class:`str`
        The ID of the community.
    name : :class:`str`
        The name of the community.
    member_count : :class:`int`
        The count of members in the community.
    is_nsfw : :class:`bool`
        Indicates if the community is NSFW.
    members_facepile_results : list[:class:`str`]
        The profile image URLs of members.
    banner : :class:`dict`
        The banner information of the community.
    is_member : :class:`bool`
        Indicates if the user is a member of the community.
    role : :class:`str`
        The role of the user in the community.
    description : :class:`str`
        The description of the community.
    creator : :class:`User` | :class:`CommunityCreator`
        The creator of the community.
    admin : :class:`User`
        The admin of the community.
    join_policy : :class:`str`
        The join policy of the community.
    created_at : :class:`int`
        The timestamp of the community's creation.
    invites_policy : :class:`str`
        The invites policy of the community.
    is_pinned : :class:`bool`
        Indicates if the community is pinned.
    rules : list[:class:`CommunityRule`]
        The rules of the community.
    """

    def __init__(self, client: Client, data: dict) -> None:
        self._client = client
        self.id: str = data['rest_id']

        self.name: str = data['name']
        self.member_count: int = data['member_count']
        self.is_nsfw: bool = data['is_nsfw']

        self.members_facepile_results: list[str] = [
            i['result']['legacy']['profile_image_url_https']
            for i in data['members_facepile_results']
        ]
        self.banner: dict = data['default_banner_media']['media_info']

        self.is_member: bool = data.get('is_member')
        self.role: str = data.get('role')
        self.description: str = data.get('description')

        if 'creator_results' in data:
            creator = data['creator_results']['result']
            if 'rest_id' in creator:
                self.creator = User(client, creator)
            else:
                self.creator = CommunityCreator(
                    b64_to_str(creator['id']).removeprefix('User:'),
                    creator['legacy']['screen_name'],
                    creator['legacy']['verified']
                )
        else:
            self.creator = None

        if 'admin_results' in data:
            admin = data['admin_results']['result']
            self.admin = User(client, admin)
        else:
            self.admin = None

        self.join_policy: str = data.get('join_policy')
        self.created_at: int = data.get('created_at')
        self.invites_policy: str = data.get('invites_policy')
        self.is_pinned: bool = data.get('is_pinned')

        if 'rules' in data:
            self.rules: list = [
                CommunityRule(i['rest_id'], i['name']) for i in data['rules']
            ]
        else:
            self.rules = None

    async def get_tweets(
        self,
        tweet_type: Literal['Top', 'Latest', 'Media'],
        count: int = 40,
        cursor: str | None = None
    ) -> Result[Tweet]:
        """
        Retrieves tweets from the community.

        Parameters
        ----------
        tweet_type : {'Top', 'Latest', 'Media'}
            The type of tweets to retrieve.
        count : :class:`int`, default=40
            The number of tweets to retrieve.

        Returns
        -------
        Result[:class:`Tweet`]
            List of retrieved tweets.

        Examples
        --------
        >>> tweets = await community.get_tweets('Latest')
        >>> for tweet in tweets:
        ...     print(tweet)
        <Tweet id="...">
        <Tweet id="...">
        ...
        >>> more_tweets = await tweets.next()  # Retrieve more tweets
        """
        return await self._client.get_community_tweets(
            self.id,
            tweet_type,
            count,
            cursor
        )

    async def join(self) -> Community:
        """
        Join the community.
        """
        return await self._client.join_community(self.id)

    async def leave(self) -> Community:
        """
        Leave the community.
        """
        return await self._client.leave_community(self.id)

    async def request_to_join(self, answer: str | None = None) -> Community:
        """
        Request to join the community.
        """
        return await self._client.request_to_join_community(self.id, answer)

    async def get_members(
        self, count: int = 20, cursor: str | None = None
    ) -> Result[CommunityMember]:
        """
        Retrieves members of the community.

        Parameters
        ----------
        count : :class:`int`, default=20
            The number of members to retrieve.

        Returns
        -------
        Result[:class:`CommunityMember`]
            List of retrieved members.
        """
        return await self._client.get_community_members(
            self.id,
            count,
            cursor
        )

    async def get_moderators(
        self, count: int = 20, cursor: str | None = None
    ) -> Result[CommunityMember]:
        """
        Retrieves moderators of the community.

        Parameters
        ----------
        count : :class:`int`, default=20
            The number of moderators to retrieve.

        Returns
        -------
        Result[:class:`CommunityMember`]
            List of retrieved moderators.
        """
        return await self._client.get_community_moderators(
            self.id,
            count,
            cursor
        )

    async def search_tweet(
        self,
        query: str,
        count: int = 20,
        cursor: str | None = None
    )-> Result[Tweet]:
        """Searchs tweets in the community.

        Parameters
        ----------
        query : :class:`str`
            The search query.
        count : :class:`int`, default=20
            The number of tweets to retrieve.

        Returns
        -------
        Result[:class:`Tweet`]
            List of retrieved tweets.
        """
        return await self._client.search_community_tweet(
            self.id,
            query,
            count,
            cursor
        )

    async def update(self) -> None:
        new = await self._client.get_community(self.id)
        self.__dict__.update(new.__dict__)

    def __eq__(self, __value: object) -> bool:
        return isinstance(__value, Community) and self.id == __value.id

    def __ne__(self, __value: object) -> bool:
        return not self == __value

    def __repr__(self) -> str:
        return f'<Community id="{self.id}">'



'''
This project includes code from the following open-source project:
https://github.com/iSarabjitDhiman/TweeterPy

We deeply appreciate the efforts of the original author.
'''

from .transaction import ClientTransaction



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

from .errors import TwitterException

if TYPE_CHECKING:
    from .client.client import Client


class Place:
    """
    Attributes
    ----------
    id : :class:`str`
        The ID of the place.
    name : :class:`str`
        The name of the place.
    full_name : :class:`str`
        The full name of the place.
    country : :class:`str`
        The country where the place is located.
    country_code : :class:`str`
        The ISO 3166-1 alpha-2 country code of the place.
    url : :class:`str`
        The URL providing more information about the place.
    place_type : :class:`str`
        The type of place.
    attributes : :class:`dict`
    bounding_box : :class:`dict`
        The bounding box that defines the geographical area of the place.
    centroid : list[:class:`float`] | None
        The geographical center of the place, represented by latitude and
        longitude.
    contained_within : list[:class:`.Place`]
        A list of places that contain this place.
    """

    def __init__(self, client: Client, data: dict) -> None:
        self._client = client

        self.id: str = data['id']
        self.name: str = data['name']
        self.full_name: str = data['full_name']
        self.country: str = data['country']
        self.country_code: str = data['country_code']
        self.url: str = data['url']
        self.place_type: str = data['place_type']
        self.attributes: dict | None = data.get('attributes')
        self.bounding_box: dict = data['bounding_box']
        self.centroid: list[float] | None = data.get('centroid')

        self.contained_within: list[Place] = [
            Place(client, place) for place in data.get('contained_within', [])
        ]

    async def update(self) -> None:
        new = self._client.get_place(self.id)
        await self.__dict__.update(new.__dict__)

    def __repr__(self) -> str:
        return f'<Place id="{self.id}" name="{self.name}">'

    def __eq__(self, __value: object) -> bool:
        return isinstance(__value, Place) and self.id == __value.id

    def __ne__(self, __value: object) -> bool:
        return not self == __value


def _places_from_response(client: Client, response: dict) -> list[Place]:
    if 'errors' in response:
        e = response['errors'][0]
        # No data available for the given coordinate.
        if e['code'] == 6:
            warnings.warn(e['message'])
        else:
            raise TwitterException(e['message'])

    places = response['result']['places'] if 'result' in response else []
    return [Place(client, place) for place in places]



from __future__ import annotations

from typing import Callable


class MockElement:
    def __init__(self, tag_name, document: MockDocument) -> None:
        self.tagName = tag_name
        self.document = document
        self.parentNode: MockElement | None = None

    def appendChild(self, child: MockElement) -> None:
        child.parentNode = self

    def remove(self) -> None:
        self.document.element_seq.remove(self)

    def removeChild(self, child: MockElement) -> None:
        child.remove()

    @property
    def lastElementChild(self) -> MockElement:
        return self.children[-1]

    def setAttribute(self, name: str, value: str) -> None:
        pass

    @property
    def children(self) -> list[MockElement]:
        return self.document._filter_elements(lambda x: x.parentNode == self)


class MockDocument:
    def __init__(self) -> None:
        self.element_seq: list[MockElement] = []
        self.createElement('body')

    def createElement(self, tag_name) -> MockElement:
        element = MockElement(tag_name, self)
        self.element_seq.append(element)
        return element

    def _filter_elements(self, function: Callable[[MockElement], list[MockElement]]) -> list[MockElement]:
        return list(
            filter(
                function,
                self.element_seq
            )
        )

    def getElementsByTagName(self, tag_name):
        return self._filter_elements(lambda x: x.tagName == tag_name)



from __future__ import annotations

import json
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..guest.client import GuestClient
    from .client import Client

    ClientType = Client | GuestClient

from ..constants import DOMAIN

class Endpoint:
    GUEST_ACTIVATE = f'https://api.{DOMAIN}/1.1/guest/activate.json'
    ONBOARDING_SSO_INIT = f'https://api.{DOMAIN}/1.1/onboarding/sso_init.json'
    ACCOUNT_LOGOUT = f'https://api.{DOMAIN}/1.1/account/logout.json'
    ONBOARDING_TASK = f'https://api.{DOMAIN}/1.1/onboarding/task.json'
    SETTINGS = f'https://api.{DOMAIN}/1.1/account/settings.json'
    UPLOAD_MEDIA = f'https://upload.{DOMAIN}/i/media/upload.json'
    UPLOAD_MEDIA_2 = f'https://upload.{DOMAIN}/i/media/upload2.json'
    CREATE_MEDIA_METADATA = f'https://api.{DOMAIN}/1.1/media/metadata/create.json'
    CREATE_CARD = f'https://caps.{DOMAIN}/v2/cards/create.json'
    VOTE = f'https://caps.{DOMAIN}/v2/capi/passthrough/1'
    REVERSE_GEOCODE = f'https://api.{DOMAIN}/1.1/geo/reverse_geocode.json'
    SEARCH_GEO = f'https://api.{DOMAIN}/1.1/geo/search.json'
    GET_PLACE = f'https://api.{DOMAIN}/1.1/geo/id/{{}}.json'
    CREATE_FRIENDSHIPS = f'https://{DOMAIN}/i/api/1.1/friendships/create.json'
    DESTROY_FRIENDSHIPS = f'https://{DOMAIN}/i/api/1.1/friendships/destroy.json'
    CREATE_BLOCKS = f'https://{DOMAIN}/i/api/1.1/blocks/create.json'
    DESTROY_BLOCKS = f'https://{DOMAIN}/i/api/1.1/blocks/destroy.json'
    CREATE_MUTES = f'https://{DOMAIN}/i/api/1.1/mutes/users/create.json'
    DESTROY_MUTES = f'https://{DOMAIN}/i/api/1.1/mutes/users/destroy.json'
    GUIDE = f'https://{DOMAIN}/i/api/2/guide.json'
    AVAILABLE_TRENDS = f'https://api.{DOMAIN}/1.1/trends/available.json'
    PLACE_TRENDS = f'https://api.{DOMAIN}/1.1/trends/place.json'
    FOLLOWERS_LIST = f'https://api.{DOMAIN}/1.1/followers/list.json'
    FRIENDS_LIST = f'https://api.{DOMAIN}/1.1/friends/list.json'
    FOLLOWERS_IDS = f'https://api.{DOMAIN}/1.1/followers/ids.json'
    FRIENDS_IDS = f'https://api.{DOMAIN}/1.1/friends/ids.json'
    DM_NEW = f'https://{DOMAIN}/i/api/1.1/dm/new2.json'
    DM_INBOX = f'https://{DOMAIN}/i/api/1.1/dm/inbox_initial_state.json'
    DM_CONVERSATION = f'https://{DOMAIN}/i/api/1.1/dm/conversation/{{}}.json'
    CONVERSATION_UPDATE_NAME = f'https://{DOMAIN}/i/api/1.1/dm/conversation/{{}}/update_name.json'
    NOTIFICATIONS_ALL = f'https://{DOMAIN}/i/api/2/notifications/all.json'
    NOTIFICATIONS_VERIFIED = f'https://{DOMAIN}/i/api/2/notifications/verified.json'
    NOTIFICATIONS_MENTIONS = f'https://{DOMAIN}/i/api/2/notifications/mentions.json'
    LIVE_PIPELINE_EVENTS = f'https://api.{DOMAIN}/live_pipeline/events'
    LIVE_PIPELINE_UPDATE_SUBSCRIPTIONS = f'https://api.{DOMAIN}/1.1/live_pipeline/update_subscriptions'
    USER_STATE = f'https://api.{DOMAIN}/help-center/forms/api/prod/user_state.json'


class V11Client:
    def __init__(self, base: ClientType) -> None:
        self.base = base

    async def guest_activate(self):
        headers = self.base._base_headers
        headers.pop('X-Twitter-Active-User', None)
        headers.pop('X-Twitter-Auth-Type', None)
        return await self.base.post(
            Endpoint.GUEST_ACTIVATE,
            headers=headers,
            data={}
        )

    async def account_logout(self):
        return await self.base.post(
            Endpoint.ACCOUNT_LOGOUT,
            headers=self.base._base_headers
        )

    async def onboarding_task(self, guest_token, token, subtask_inputs, data = None, **kwargs):
        if data is None:
            data = {}
        if token is not None:
            data['flow_token'] = token
        if subtask_inputs is not None:
            data['subtask_inputs'] = subtask_inputs

        headers = {
            'x-guest-token': guest_token,
            'Authorization': 'Bearer AAAAAAAAAAAAAAAAAAAAANRILgAAAAAAnNwIzUejRCOuH5E6I8xnZz4puTs%3D1Zv7ttfk8LF81IUq16cHjhLTvJu4FA33AGWWjCpTnA'
        }

        if self.base._get_csrf_token():
            headers["x-csrf-token"] = self.base._get_csrf_token()
            headers["x-twitter-auth-type"] = "OAuth2Session"

        return await self.base.post(
            Endpoint.ONBOARDING_TASK,
            json=data,
            headers=headers,
            **kwargs
        )

    async def sso_init(self, provider, guest_token):
        headers = self.base._base_headers | {
            'x-guest-token': guest_token
        }
        headers.pop('X-Twitter-Active-User')
        headers.pop('X-Twitter-Auth-Type')
        return await self.base.post(
            Endpoint.ONBOARDING_SSO_INIT,
            json={'provider': provider},
            headers=headers
        )

    async def settings(self):
        return await self.base.get(
            Endpoint.SETTINGS,
            headers=self.base._base_headers
        )

    async def upload_media(self, method, is_long_video: bool, *args, **kwargs):
        if is_long_video:
            endpoint = Endpoint.UPLOAD_MEDIA_2
        else:
            endpoint = Endpoint.UPLOAD_MEDIA
        return await self.base.request(method, endpoint, *args, **kwargs)

    async def upload_media_init(self, media_type, total_bytes, media_category, is_long_video: bool):
        params = {
            'command': 'INIT',
            'total_bytes': total_bytes,
            'media_type': media_type
        }
        if media_category is not None:
            params['media_category'] = media_category

        return await self.upload_media(
            'POST',
            is_long_video,
            params=params,
            headers=self.base._base_headers
        )

    async def upload_media_append(self, is_long_video, media_id, segment_index, chunk_stream):
        params = {
            'command': 'APPEND',
            'media_id': media_id,
            'segment_index': segment_index,
        }
        headers = self.base._base_headers
        headers.pop('content-type')
        files = {
            'media': (
                'blob',
                chunk_stream,
                'application/octet-stream',
            )
        }
        return await self.upload_media(
            'POST',
            is_long_video,
            params=params,
            headers=headers, files=files
        )

    async def upload_media_finelize(self, is_long_video, media_id):
        params = {
            'command': 'FINALIZE',
            'media_id': media_id,
        }
        return await self.upload_media(
            'POST',
            is_long_video,
            params=params,
            headers=self.base._base_headers,
        )

    async def upload_media_status(self, is_long_video, media_id):
        params = {
            'command': 'STATUS',
            'media_id': media_id,
        }
        return await self.upload_media(
            'GET',
            is_long_video,
            params=params,
            headers=self.base._base_headers,
        )

    async def create_media_metadata(self, media_id, alt_text, sensitive_warning):
        data = {'media_id': media_id}
        if alt_text is not None:
            data['alt_text'] = {'text': alt_text}
        if sensitive_warning is not None:
            data['sensitive_media_warning'] = sensitive_warning
        return await self.base.post(
            Endpoint.CREATE_MEDIA_METADATA,
            json=data,
            headers=self.base._base_headers
        )

    async def create_card(self, choices, duration_minutes):
        card_data = {
            'twitter:card': f'poll{len(choices)}choice_text_only',
            'twitter:api:api:endpoint': '1',
            'twitter:long:duration_minutes': duration_minutes
        }

        for i, choice in enumerate(choices, 1):
            card_data[f'twitter:string:choice{i}_label'] = choice

        data = {'card_data': json.dumps(card_data)}
        headers = self.base._base_headers | {'content-type': 'application/x-www-form-urlencoded'}
        return await self.base.post(
            Endpoint.CREATE_CARD,
            data=data,
            headers=headers,
        )

    async def vote(self, selected_choice: str, card_uri: str, tweet_id: str, card_name: str):
        data = {
            'twitter:string:card_uri': card_uri,
            'twitter:long:original_tweet_id': tweet_id,
            'twitter:string:response_card_name': card_name,
            'twitter:string:cards_platform': 'Web-12',
            'twitter:string:selected_choice': selected_choice
        }
        headers = self.base._base_headers | {
            'content-type': 'application/x-www-form-urlencoded'
        }
        return await self.base.post(
            Endpoint.VOTE,
            data=data,
            headers=headers
        )

    async def reverse_geocode(self, lat, long, accuracy, granularity, max_results):
        params = {
            'lat': lat,
            'long': long,
            'accuracy': accuracy,
            'granularity': granularity,
            'max_results': max_results
        }
        for k, v in tuple(params.items()):
            if v is None:
                params.pop(k)
        return await self.base.get(
            Endpoint.REVERSE_GEOCODE,
            params=params,
            headers=self.base._base_headers
        )

    async def search_geo(self, lat, long, query, ip, granularity, max_results):
        params = {
            'lat': lat,
            'long': long,
            'query': query,
            'ip': ip,
            'granularity': granularity,
            'max_results': max_results
        }
        for k, v in tuple(params.items()):
            if v is None:
                params.pop(k)

        return await self.base.get(
            Endpoint.SEARCH_GEO,
            params=params,
            headers=self.base._base_headers
        )

    async def get_place(self, id):
        return await self.base.get(
            Endpoint.GET_PLACE.format(id),
            headers=self.base._base_headers
        )

    async def create_friendships(self, user_id):
        data = {
            'include_profile_interstitial_type': 1,
            'include_blocking': 1,
            'include_blocked_by': 1,
            'include_followed_by': 1,
            'include_want_retweets': 1,
            'include_mute_edge': 1,
            'include_can_dm': 1,
            'include_can_media_tag': 1,
            'include_ext_is_blue_verified': 1,
            'include_ext_verified_type': 1,
            'include_ext_profile_image_shape': 1,
            'skip_status': 1,
            'user_id': user_id
        }
        headers = self.base._base_headers | {
            'content-type': 'application/x-www-form-urlencoded'
        }
        return await self.base.post(
            Endpoint.CREATE_FRIENDSHIPS,
            data=data,
            headers=headers
        )

    async def destroy_friendships(self, user_id):
        data = {
            'include_profile_interstitial_type': 1,
            'include_blocking': 1,
            'include_blocked_by': 1,
            'include_followed_by': 1,
            'include_want_retweets': 1,
            'include_mute_edge': 1,
            'include_can_dm': 1,
            'include_can_media_tag': 1,
            'include_ext_is_blue_verified': 1,
            'include_ext_verified_type': 1,
            'include_ext_profile_image_shape': 1,
            'skip_status': 1,
            'user_id': user_id
        }
        headers = self.base._base_headers | {
            'content-type': 'application/x-www-form-urlencoded'
        }
        return await self.base.post(
            Endpoint.DESTROY_FRIENDSHIPS,
            data=data,
            headers=headers
        )

    async def create_blocks(self, user_id):
        data = {'user_id': user_id}
        headers = self.base._base_headers
        headers['content-type'] = 'application/x-www-form-urlencoded'
        return await self.base.post(
            Endpoint.CREATE_BLOCKS,
            data=data,
            headers=headers
        )

    async def destroy_blocks(self, user_id):
        data = {'user_id': user_id}
        headers = self.base._base_headers
        headers['content-type'] = 'application/x-www-form-urlencoded'
        return await self.base.post(
            Endpoint.DESTROY_BLOCKS,
            data=data,
            headers=headers
        )

    async def create_mutes(self, user_id):
        data = {'user_id': user_id}
        headers = self.base._base_headers
        headers['content-type'] = 'application/x-www-form-urlencoded'
        return await self.base.post(
            Endpoint.CREATE_MUTES,
            data=data,
            headers=headers
        )

    async def destroy_mutes(self, user_id):
        data = {'user_id': user_id}
        headers = self.base._base_headers
        headers['content-type'] = 'application/x-www-form-urlencoded'
        return await self.base.post(
            Endpoint.DESTROY_MUTES,
            data=data,
            headers=headers
        )

    async def guide(self, category, count, additional_request_params):
        params = {
            'count': count,
            'include_page_configuration': True,
            'initial_tab_id': category
        }
        if additional_request_params is not None:
            params |= additional_request_params
        return await self.base.get(
            Endpoint.GUIDE,
            params=params,
            headers=self.base._base_headers
        )

    async def available_trends(self):
        return await self.base.get(
            Endpoint.AVAILABLE_TRENDS,
            headers=self.base._base_headers
        )

    async def place_trends(self, woeid):
        return await self.base.get(
            Endpoint.PLACE_TRENDS,
            params={'id': woeid},
            headers=self.base._base_headers
        )

    async def _friendships(self, user_id, screen_name, count, endpoint, cursor):
        params = {'count': count}
        if user_id is not None:
            params['user_id'] = user_id
        elif screen_name is not None:
            params['screen_name'] = screen_name

        if cursor is not None:
            params['cursor'] = cursor

        return await self.base.get(
            endpoint,
            params=params,
            headers=self.base._base_headers
        )

    async def followers_list(self, user_id, screen_name, count, cursor):
        return await self._friendships(user_id, screen_name, count, Endpoint.FOLLOWERS_LIST, cursor)

    async def friends_list(self, user_id, screen_name, count, cursor):
        return await self._friendships(user_id, screen_name, count, Endpoint.FRIENDS_LIST, cursor)

    async def _friendship_ids(self, user_id, screen_name, count, endpoint, cursor):
        params = {'count': count}
        if user_id is not None:
            params['user_id'] = user_id
        elif user_id is not None:
            params['screen_name'] = screen_name

        if cursor is not None:
            params['cursor'] = cursor

        return await self.base.get(
            endpoint,
            params=params,
            headers=self.base._base_headers
        )

    async def followers_ids(self, user_id, screen_name, count, cursor):
        return await self._friendship_ids(user_id, screen_name, count, Endpoint.FOLLOWERS_IDS, cursor)

    async def friends_ids(self, user_id, screen_name, count, cursor):
        return await self._friendship_ids(user_id, screen_name, count, Endpoint.FRIENDS_IDS, cursor)

    async def dm_new(self, conversation_id, text, media_id, reply_to):
        data = {
            'cards_platform': 'Web-12',
            'conversation_id': conversation_id,
            'dm_users': False,
            'include_cards': 1,
            'include_quote_count': True,
            'recipient_ids': False,
            'text': text
        }
        if media_id is not None:
            data['media_id'] = media_id
        if reply_to is not None:
            data['reply_to_dm_id'] = reply_to

        return await self.base.post(
            Endpoint.DM_NEW,
            json=data,
            headers=self.base._base_headers
        )

    async def dm_conversation(self, conversation_id, max_id):
        params = {'context': 'FETCH_DM_CONVERSATION_HISTORY', 'include_conversation_info': True}
        if max_id is not None:
            params['max_id'] = max_id

        return await self.base.get(
            Endpoint.DM_CONVERSATION.format(conversation_id),
            params=params,
            headers=self.base._base_headers
        )

    async def conversation_update_name(self, group_id, name):
        data = {'name': name}
        headers = self.base._base_headers
        headers['content-type'] = 'application/x-www-form-urlencoded'
        return await self.base.post(
            Endpoint.CONVERSATION_UPDATE_NAME.format(group_id),
            data=data,
            headers=headers
        )

    async def _notifications(self, endpoint, count, cursor):
        params = {'count': count}
        if cursor is not None:
            params['cursor'] = cursor

        return await self.base.get(
            endpoint,
            params=params,
            headers=self.base._base_headers
        )

    async def notifications_all(self, count, cursor):
        return await self._notifications(Endpoint.NOTIFICATIONS_ALL, count, cursor)

    async def notifications_verified(self, count, cursor):
        return await self._notifications(Endpoint.NOTIFICATIONS_VERIFIED, count, cursor)

    async def notifications_mentions(self, count, cursor):
        return await self._notifications(Endpoint.NOTIFICATIONS_MENTIONS, count, cursor)

    async def live_pipeline_update_subscriptions(self, session, subscribe, unsubscribe):
        data = {
            'sub_topics': subscribe,
            'unsub_topics': unsubscribe
        }
        headers = self.base._base_headers
        headers['content-type'] = 'application/x-www-form-urlencoded'
        headers['LivePipeline-Session'] = session
        return await self.base.post(
            Endpoint.LIVE_PIPELINE_UPDATE_SUBSCRIPTIONS, data=data, headers=headers
        )

    async def user_state(self):
        return await self.base.get(
            Endpoint.USER_STATE,
            headers=self.base._base_headers
        )



from __future__ import annotations

import json
import warnings
from functools import partial
from typing import Any, Literal
from urllib.parse import urlparse

from httpx import AsyncClient, AsyncHTTPTransport, Response
from httpx._utils import URLPattern

from ..client.gql import GQLClient
from ..client.v11 import V11Client
from ..constants import DOMAIN, TOKEN
from ..errors import (
    BadRequest,
    Forbidden,
    NotFound,
    RequestTimeout,
    ServerError,
    TooManyRequests,
    TwitterException,
    Unauthorized
)
from ..utils import Result, find_dict, find_entry_by_type, httpx_transport_to_url
from ..x_client_transaction import ClientTransaction
from .tweet import Tweet
from .user import User


def tweet_from_data(client: GuestClient, data: dict, index: int = 0) -> Tweet:
    ':meta private:'
    tweet_data_ = find_dict(data, 'result', index == 0)
    if tweet_data_ and index > 0:
        tweet_data_ = [ _i for _i in tweet_data_ if _i.get('__typename') == 'Tweet']
    if not tweet_data_ or len(tweet_data_) <= index:
        return None
    tweet_data = tweet_data_[index]

    if tweet_data.get('__typename') == 'TweetTombstone':
        return None
    if 'tweet' in tweet_data:
        tweet_data = tweet_data['tweet']
    if 'core' not in tweet_data:
        return None
    if 'result' not in tweet_data['core']['user_results']:
        return None
    if 'legacy' not in tweet_data:
        return None

    user_data = tweet_data['core']['user_results']['result']
    return Tweet(client, tweet_data, User(client, user_data))


class GuestClient:
    """
    A client for interacting with the Twitter API as a guest.
    This class is used for interacting with the Twitter API
    without requiring authentication.

    Parameters
    ----------
    language : :class:`str` | None, default=None
        The language code to use in API requests.
    proxy : :class:`str` | None, default=None
        The proxy server URL to use for request
        (e.g., 'http://0.0.0.0:0000').

    Examples
    --------
    >>> client = GuestClient()
    >>> await client.activate()  # Activate the client by generating a guest token.
    """

    def __init__(
        self,
        language: str = 'en-US',
        proxy: str | None = None,
        **kwargs
    ) -> None:
        if 'proxies' in kwargs:
            message = (
                "The 'proxies' argument is now deprecated. Use 'proxy' "
                "instead. https://github.com/encode/httpx/pull/2879"
            )
            warnings.warn(message)

        self.http = AsyncClient(proxy=proxy, **kwargs)
        self.language = language
        self.proxy = proxy

        self._token = TOKEN
        self._user_agent = ('Mozilla/5.0 (Windows NT 10.0; Win64; x64) '
                            'AppleWebKit/537.36 (KHTML, like Gecko) '
                            'Chrome/122.0.0.0 Safari/537.36')
        self._guest_token: str | None = None  # set when activate method is called
        self.gql = GQLClient(self)
        self.v11 = V11Client(self)
        self.client_transaction = ClientTransaction()

    async def request(
        self,
        method: str,
        url: str,
        raise_exception: bool = True,
        **kwargs
    ) -> tuple[dict | Any, Response]:
        ':meta private:'
        headers = kwargs.pop('headers', {})

        if not self.client_transaction.home_page_response:
            cookies_backup = dict(self.http.cookies).copy()
            ct_headers = {
                'Accept-Language': f'{self.language},{self.language.split("-")[0]};q=0.9',
                'Cache-Control': 'no-cache',
                'Referer': f'https://{DOMAIN}',
                'User-Agent': self._user_agent
            }
            await self.client_transaction.init(self.http, ct_headers)
            self.http.cookies = cookies_backup

        tid = self.client_transaction.generate_transaction_id(method=method, path=urlparse(url).path)
        headers['X-Client-Transaction-Id'] = tid

        response = await self.http.request(method, url, headers=headers, **kwargs)

        try:
            response_data = response.json()
        except json.decoder.JSONDecodeError:
            response_data = response.text

        status_code = response.status_code

        if status_code >= 400 and raise_exception:
            message = f'status: {status_code}, message: "{response.text}"'
            if status_code == 400:
                raise BadRequest(message, headers=response.headers)
            elif status_code == 401:
                raise Unauthorized(message, headers=response.headers)
            elif status_code == 403:
                raise Forbidden(message, headers=response.headers)
            elif status_code == 404:
                raise NotFound(message, headers=response.headers)
            elif status_code == 408:
                raise RequestTimeout(message, headers=response.headers)
            elif status_code == 429:
                raise TooManyRequests(message, headers=response.headers)
            elif 500 <= status_code < 600:
                raise ServerError(message, headers=response.headers)
            else:
                raise TwitterException(message, headers=response.headers)

        return response_data, response

    async def get(self, url, **kwargs) -> tuple[dict | Any, Response]:
        ':meta private:'
        return await self.request('GET', url, **kwargs)

    async def post(self, url, **kwargs) -> tuple[dict | Any, Response]:
        ':meta private:'
        return await self.request('POST', url, **kwargs)

    @property
    def proxy(self) -> str:
        ':meta private:'
        transport: AsyncHTTPTransport = self.http._mounts.get(
            URLPattern('all://')
        )
        if transport is None:
            return None
        if not hasattr(transport._pool, '_proxy_url'):
            return None
        return httpx_transport_to_url(transport)

    @proxy.setter
    def proxy(self, url: str) -> None:
        self.http._mounts = {
            URLPattern('all://'): AsyncHTTPTransport(proxy=url)
        }

    @property
    def _base_headers(self) -> dict[str, str]:
        """
        Base headers for Twitter API requests.
        """
        headers = {
            'authorization': f'Bearer {self._token}',
            'content-type': 'application/json',
            'X-Twitter-Active-User': 'yes',
            'Referer': f'https://{DOMAIN}',
        }

        if self.language is not None:
            headers['Accept-Language'] = self.language
            headers['X-Twitter-Client-Language'] = self.language

        if self._guest_token is not None:
            headers['X-Guest-Token'] = self._guest_token

        return headers

    async def activate(self) -> str:
        """
        Activate the client by generating a guest token.
        """
        response, _ = await self.v11.guest_activate()
        self._guest_token = response['guest_token']
        return self._guest_token

    async def get_user_by_screen_name(self, screen_name: str) -> User:
        """
        Retrieves a user object based on the provided screen name.

        Parameters
        ----------
        screen_name : :class:`str`
            The screen name of the user to retrieve.

        Returns
        -------
        :class:`.user.User`
            An instance of the `User` class containing user details.

        Examples
        --------
        >>> user = await client.get_user_by_screen_name('example_user')
        >>> print(user)
        <User id="...">
        """
        response, _ = await self.gql.user_by_screen_name(screen_name)
        return User(self, response['data']['user']['result'])

    async def get_user_by_id(self, user_id: str) -> User:
        """
        Retrieves a user object based on the provided user ID.

        Parameters
        ----------
        user_id : :class:`str`
            The ID of the user to retrieve.

        Returns
        -------
        :class:`.user.User`
            An instance of the `User` class

        Examples
        --------
        >>> user = await client.get_user_by_id('123456789')
        >>> print(user)
        <User id="123456789">
        """
        response, _ = await self.gql.user_by_rest_id(user_id)
        return User(self, response['data']['user']['result'])

    async def get_user_tweets(
        self,
        user_id: str,
        tweet_type: Literal['Tweets'] = 'Tweets',
        count: int = 40,
    ) -> list[Tweet]:
        """
        Fetches tweets from a specific user's timeline.

        Parameters
        ----------
        user_id : :class:`str`
            The ID of the Twitter user whose tweets to retrieve.
            To get the user id from the screen name, you can use
            `get_user_by_screen_name` method.
        tweet_type : {'Tweets'}, default='Tweets'
            The type of tweets to retrieve.
        count : :class:`int`, default=40
            The number of tweets to retrieve.

        Returns
        -------
        list[:class:`.tweet.Tweet`]
            A Result object containing a list of `Tweet` objects.

        Examples
        --------
        >>> user_id = '...'

        If you only have the screen name, you can get the user id as follows:

        >>> screen_name = 'example_user'
        >>> user = client.get_user_by_screen_name(screen_name)
        >>> user_id = user.id

        >>> tweets = await client.get_user_tweets(user_id)
        >>> for tweet in tweets:
        ...    print(tweet)
        <Tweet id="...">
        <Tweet id="...">
        ...
        ...

        See Also
        --------
        .get_user_by_screen_name
        """
        tweet_type = tweet_type.capitalize()
        f = {
            'Tweets': self.gql.user_tweets,
        }[tweet_type]
        response, _ = await f(user_id, count, None)
        instructions_ = find_dict(response, 'instructions', True)
        if not instructions_:
            return []
        instructions = instructions_[0]
        items = find_entry_by_type(instructions, 'TimelineAddEntries')['entries']
        results = []

        for item in items:
            entry_id = item['entryId']
            if not entry_id.startswith(('tweet', 'profile-conversation', 'profile-grid')):
                continue
            if entry_id.startswith('profile-conversation'):
                items_ = find_dict(response, 'items', True)[0]
                for _index in range(len(items_)):
                    tweet = tweet_from_data(self, item, _index)
                    if tweet is not None:
                        results.append(tweet)
            else:
                tweet = tweet_from_data(self, item)
                if tweet is not None:
                    results.append(tweet)

        return results

    async def get_tweet_by_id(self, tweet_id: str) -> Tweet:
        """
        Fetches a tweet by tweet ID.

        Parameters
        ----------
        tweet_id : :class:`str`
            The ID of the tweet.

        Returns
        -------
        :class:`.tweet.Tweet`
            Tweet object

        Examples
        --------
        >>> await client.get_tweet_by_id('123456789')
        <Tweet id="123456789">
        """
        response, _ = await self.gql.tweet_result_by_rest_id(tweet_id)
        return tweet_from_data(self, response)

    async def get_user_highlights_tweets(
        self,
        user_id: str,
        count: int = 20,
        cursor: str | None = None
    ) -> Result[Tweet]:
        """
        Retrieves highlighted tweets from a user's timeline.

        Parameters
        ----------
        user_id : :class:`str`
            The user ID
        count : :class:`int`, default=20
            The number of tweets to retrieve.

        Returns
        -------
        Result[:class:`.tweet.Tweet`]
            An instance of the `Result` class containing the highlighted tweets.

        Examples
        --------
        >>> result = await client.get_user_highlights_tweets('123456789')
        >>> for tweet in result:
        ...     print(tweet)
        <Tweet id="...">
        <Tweet id="...">
        ...
        ...

        >>> more_results = await result.next()  # Retrieve more highlighted tweets
        >>> for tweet in more_results:
        ...     print(tweet)
        <Tweet id="...">
        <Tweet id="...">
        ...
        ...
        """
        response, _ = await self.gql.user_highlights_tweets(user_id, count, cursor)

        instructions = response['data']['user']['result']['timeline']['timeline']['instructions']
        instruction = find_entry_by_type(instructions, 'TimelineAddEntries')
        if instruction is None:
            return Result.empty()
        entries = instruction['entries']
        previous_cursor = None
        next_cursor = None
        results = []

        for entry in entries:
            entryId = entry['entryId']
            if entryId.startswith('tweet'):
                results.append(tweet_from_data(self, entry))
            elif entryId.startswith('cursor-top'):
                previous_cursor = entry['content']['value']
            elif entryId.startswith('cursor-bottom'):
                next_cursor = entry['content']['value']

        return Result(
            results,
            partial(self.get_user_highlights_tweets, user_id, count, next_cursor),
            next_cursor,
            partial(self.get_user_highlights_tweets, user_id, count, previous_cursor),
            previous_cursor
        )



from typing import Union, List


def interpolate(from_list: List[Union[float, int]], to_list: List[Union[float, int]], f: Union[float, int]):
    if len(from_list) != len(to_list):
        raise Exception(
            f"Mismatched interpolation arguments {from_list}: {to_list}")
    out = []
    for i in range(len(from_list)):
        out.append(interpolate_num(from_list[i], to_list[i], f))
    return out


def interpolate_num(from_val: List[Union[float, int]], to_val: List[Union[float, int]], f: Union[float, int]):
    if all([isinstance(number, (int, float)) for number in [from_val, to_val]]):
        return from_val * (1 - f) + to_val * f

    if all([isinstance(number, bool) for number in [from_val, to_val]]):
        return from_val if f < 0.5 else to_val


if __name__ == "__main__":
    pass



import re
import bs4
import base64
from typing import Union


async def handle_x_migration(session, headers):
    home_page = None
    migration_redirection_regex = re.compile(
        r"""(http(?:s)?://(?:www\.)?(twitter|x){1}\.com(/x)?/migrate([/?])?tok=[a-zA-Z0-9%\-_]+)+""", re.VERBOSE)
    response = await session.request(method="GET", url="https://x.com", headers=headers)
    home_page = bs4.BeautifulSoup(response.content, 'lxml')
    migration_url = home_page.select_one("meta[http-equiv='refresh']")
    migration_redirection_url = re.search(migration_redirection_regex, str(
        migration_url)) or re.search(migration_redirection_regex, str(response.content))
    if migration_redirection_url:
        response = await session.request(method="GET", url=migration_redirection_url.group(0), headers=headers)
        home_page = bs4.BeautifulSoup(response.content, 'lxml')
    migration_form = home_page.select_one("form[name='f']") or home_page.select_one(f"form[action='https://x.com/x/migrate']")
    if migration_form:
        url = migration_form.attrs.get("action", "https://x.com/x/migrate") + "/?mx=2"
        method = migration_form.attrs.get("method", "POST")
        request_payload = {input_field.get("name"): input_field.get("value") for input_field in migration_form.select("input")}
        response = await session.request(method=method, url=url, data=request_payload, headers=headers)
        home_page = bs4.BeautifulSoup(response.content, 'lxml')
    return home_page


def float_to_hex(x):
    result = []
    quotient = int(x)
    fraction = x - quotient

    while quotient > 0:
        quotient = int(x / 16)
        remainder = int(x - (float(quotient) * 16))

        if remainder > 9:
            result.insert(0, chr(remainder + 55))
        else:
            result.insert(0, str(remainder))

        x = float(quotient)

    if fraction == 0:
        return ''.join(result)

    result.append('.')

    while fraction > 0:
        fraction *= 16
        integer = int(fraction)
        fraction -= float(integer)

        if integer > 9:
            result.append(chr(integer + 55))
        else:
            result.append(str(integer))

    return ''.join(result)


def is_odd(num: Union[int, float]):
    if num % 2:
        return -1.0
    return 0.0


def base64_encode(string):
    string = string.encode() if isinstance(string, str) else string
    return base64.b64encode(string).decode()


def base64_decode(input):
    try:
        data = base64.b64decode(input)
        return data.decode()
    except Exception:
        # return bytes(input, "utf-8")
        return list(bytes(input, "utf-8"))


if __name__ == "__main__":
    pass



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Literal

from .utils import timestamp_to_datetime

if TYPE_CHECKING:
    from httpx import Response

    from .client.client import Client
    from .tweet import Tweet
    from .user import User
    from .utils import Result


class List:
    """
    Class representing a Twitter List.

    Attributes
    ----------
    id : :class:`str`
        The unique identifier of the List.
    created_at : :class:`int`
        The timestamp when the List was created.
    default_banner : :class:`dict`
        Information about the default banner of the List.
    banner : :class:`dict`
        Information about the banner of the List. If custom banner is not set,
        it defaults to the default banner.
    description : :class:`str`
        The description of the List.
    following : :class:`bool`
        Indicates if the authenticated user is following the List.
    is_member : :class:`bool`
        Indicates if the authenticated user is a member of the List.
    member_count : :class:`int`
        The number of members in the List.
    mode : {'Private', 'Public'}
        The mode of the List, either 'Private' or 'Public'.
    muting : :class:`bool`
        Indicates if the authenticated user is muting the List.
    name : :class:`str`
        The name of the List.
    pinning : :class:`bool`
        Indicates if the List is pinned.
    subscriber_count : :class:`int`
        The number of subscribers to the List.
    """
    def __init__(self, client: Client, data: dict) -> None:
        self._client = client

        self.id: str = data['id_str']
        self.created_at: int = data['created_at']
        self.default_banner: dict = data['default_banner_media']['media_info']

        if 'custom_banner_media' in data:
            self.banner: dict = data["custom_banner_media"]["media_info"]
        else:
            self.banner: dict = self.default_banner

        self.description: str = data['description']
        self.following: bool = data['following']
        self.is_member: bool = data['is_member']
        self.member_count: bool = data['member_count']
        self.mode: Literal['Private', 'Public'] = data['mode']
        self.muting: bool = data['muting']
        self.name: str = data['name']
        self.pinning: bool = data['pinning']
        self.subscriber_count: int = data['subscriber_count']

    @property
    def created_at_datetime(self) -> datetime:
        return timestamp_to_datetime(self.created_at)

    async def edit_banner(self, media_id: str) -> Response:
        """
        Edit the banner image of the list.

        Parameters
        ----------
        media_id : :class:`str`
            The ID of the media to use as the new banner image.

        Returns
        -------
        :class:`httpx.Response`
            Response returned from twitter api.

        Examples
        --------
        >>> media_id = await client.upload_media('image.png')
        >>> await media.edit_banner(media_id)
        """
        return await self._client.edit_list_banner(self.id, media_id)

    async def delete_banner(self) -> Response:
        """
        Deletes the list banner.
        """
        return await self._client.delete_list_banner(self.id)

    async def edit(
        self,
        name: str | None = None,
        description: str | None = None,
        is_private: bool | None = None
    ) -> List:
        """
        Edits list information.

        Parameters
        ----------
        name : :class:`str`, default=None
            The new name for the list.
        description : :class:`str`, default=None
            The new description for the list.
        is_private : :class:`bool`, default=None
            Indicates whether the list should be private
            (True) or public (False).

        Returns
        -------
        :class:`List`
            The updated Twitter list.

        Examples
        --------
        >>> await list.edit(
        ...     'new name', 'new description', True
        ... )
        """
        return await self._client.edit_list(
            self.id, name, description, is_private
        )

    async def add_member(self, user_id: str) -> Response:
        """
        Adds a member to the list.
        """
        return await self._client.add_list_member(self.id, user_id)

    async def remove_member(self, user_id: str) -> Response:
        """
        Removes a member from the list.
        """
        return await self._client.remove_list_member(self.id, user_id)

    async def get_tweets(
        self, count: int = 20, cursor: str | None = None
    ) -> Result[Tweet]:
        """
        Retrieves tweets from the list.

        Parameters
        ----------
        count : :class:`int`, default=20
            The number of tweets to retrieve.
        cursor : :class:`str`, default=None
            The cursor for pagination.

        Returns
        -------
        Result[:class:`Tweet`]
            A Result object containing the retrieved tweets.

        Examples
        --------
        >>> tweets = await list.get_tweets()
        >>> for tweet in tweets:
        ...    print(tweet)
        <Tweet id="...">
        <Tweet id="...">
        ...
        ...

        >>> more_tweets = await tweets.next()  # Retrieve more tweets
        >>> for tweet in more_tweets:
        ...     print(tweet)
        <Tweet id="...">
        <Tweet id="...">
        ...
        ...
        """
        return await self._client.get_list_tweets(self.id, count, cursor)

    async def get_members(
        self, count: int = 20, cursor: str | None = None
    ) -> Result[User]:
        """Retrieves members of the list.

        Parameters
        ----------
        count : :class:`int`, default=20
            Number of members to retrieve.

        Returns
        -------
        Result[:class:`User`]
            Members of the list

        Examples
        --------
        >>> members = list_.get_members()
        >>> for member in members:
        ...     print(member)
        <User id="...">
        <User id="...">
        ...
        ...
        >>> more_members = members.next()  # Retrieve more members
        """
        return await self._client.get_list_members(self.id, count, cursor)

    async def get_subscribers(
        self, count: int = 20, cursor: str | None = None
    ) -> Result[User]:
        """Retrieves subscribers of the list.

        Parameters
        ----------
        count : :class:`int`, default=20
            Number of subscribers to retrieve.

        Returns
        -------
        Result[:class:`User`]
            Subscribers of the list

        Examples
        --------
        >>> subscribers = list_.get_subscribers()
        >>> for subscriber in subscribers:
        ...     print(subscriber)
        <User id="...">
        <User id="...">
        ...
        ...
        >>> more_subscribers = subscribers.next()  # Retrieve more subscribers
        """
        return await self._client.get_list_subscribers(self.id, count, cursor)

    async def update(self) -> None:
        new = await self._client.get_list(self.id)
        self.__dict__.update(new.__dict__)

    def __eq__(self, __value: object) -> bool:
        return isinstance(__value, List) and self.id == __value.id

    def __ne__(self, __value: object) -> bool:
        return not self == __value

    def __repr__(self) -> str:
        return f'<List id="{self.id}">'



from __future__ import annotations

import base64
import json
from datetime import datetime
from httpx import AsyncHTTPTransport
from typing import TYPE_CHECKING, Any, Awaitable, Generic, Iterator, Literal, TypedDict, TypeVar

if TYPE_CHECKING:
    from .client.client import Client

T = TypeVar('T')


class Result(Generic[T]):
    """
    This class is for storing multiple results.
    The `next` method can be used to retrieve further results.
    As with a regular list, you can access elements by
    specifying indexes and iterate over elements using a for loop.

    Attributes
    ----------
    next_cursor : :class:`str`
        Cursor used to obtain the next result.
    previous_cursor : :class:`str`
        Cursor used to obtain the previous result.
    token : :class:`str`
        Alias of `next_cursor`.
    cursor : :class:`str`
        Alias of `next_cursor`.
    """

    def __init__(
        self,
        results: list[T],
        fetch_next_result: Awaitable | None = None,
        next_cursor: str | None = None,
        fetch_previous_result: Awaitable | None = None,
        previous_cursor: str | None = None
    ) -> None:
        self.__results = results
        self.next_cursor = next_cursor
        self.__fetch_next_result = fetch_next_result
        self.previous_cursor = previous_cursor
        self.__fetch_previous_result = fetch_previous_result

    async def next(self) -> Result[T]:
        """
        The next result.
        """
        if self.__fetch_next_result is None:
            return Result([])
        return await self.__fetch_next_result()

    async def previous(self) -> Result[T]:
        """
        The previous result.
        """
        if self.__fetch_previous_result is None:
            return Result([])
        return await self.__fetch_previous_result()

    @classmethod
    def empty(cls):
        return cls([])

    def __iter__(self) -> Iterator[T]:
        yield from self.__results

    def __getitem__(self, index: int) -> T:
        return self.__results[index]

    def __len__(self) -> int:
        return len(self.__results)

    def __repr__(self) -> str:
        return self.__results.__repr__()


class Flow:
    def __init__(self, client: Client, guest_token: str) -> None:
        self._client = client
        self.guest_token = guest_token
        self.response = None

    async def execute_task(self, *subtask_inputs, **kwargs) -> None:
        response, _ = await self._client.v11.onboarding_task(
            self.guest_token, self.token, list(subtask_inputs), **kwargs
        )
        self.response = response

    async def sso_init(self, provider: str) -> None:
        await self._client.v11.sso_init(provider, self.guest_token)

    @property
    def token(self) -> str | None:
        if self.response is None:
            return None
        return self.response.get('flow_token')

    @property
    def task_id(self) -> str | None:
        if self.response is None:
            return None
        if len(self.response['subtasks']) <= 0:
            return None
        return self.response['subtasks'][0]['subtask_id']


def find_dict(obj: list | dict, key: str | int, find_one: bool = False) -> list[Any]:
    """
    Retrieves elements from a nested dictionary.
    """
    results = []
    if isinstance(obj, dict):
        if key in obj:
            results.append(obj.get(key))
            if find_one:
                return results
    if isinstance(obj, (list, dict)):
        for elem in (obj if isinstance(obj, list) else obj.values()):
            r = find_dict(elem, key, find_one)
            results += r
            if r and find_one:
                return results
    return results


def httpx_transport_to_url(transport: AsyncHTTPTransport) -> str:
    url = transport._pool._proxy_url
    scheme = url.scheme.decode()
    host = url.host.decode()
    port = url.port
    auth = None
    if transport._pool._proxy_headers:
        auth_header = dict(transport._pool._proxy_headers)[b'Proxy-Authorization'].decode()
        auth = base64.b64decode(auth_header.split()[1]).decode()

    url_str = f'{scheme}://'
    if auth is not None:
        url_str += auth + '@'
    url_str += host
    if port is not None:
        url_str += f':{port}'
    return url_str


def get_query_id(url: str) -> str:
    """
    Extracts the identifier from a URL.

    Examples
    --------
    >>> get_query_id('https://twitter.com/i/api/graphql/queryid/...')
    'queryid'
    """
    return url.rsplit('/', 2)[-2]


def timestamp_to_datetime(timestamp: str) -> datetime:
    return datetime.strptime(timestamp, '%a %b %d %H:%M:%S %z %Y')


def build_tweet_data(raw_data: dict) -> dict:
    return {
        **raw_data,
        'rest_id': raw_data['id'],
        'is_translatable': None,
        'views': {},
        'edit_control': {},
        'legacy': {
            'created_at': raw_data.get('created_at'),
            'full_text': raw_data.get('full_text') or raw_data.get('text'),
            'lang': raw_data.get('lang'),
            'is_quote_status': raw_data.get('is_quote_status'),
            'in_reply_to_status_id_str': raw_data.get('in_reply_to_status_id_str'),
            'retweeted_status_result': raw_data.get('retweeted_status_result'),
            'possibly_sensitive': raw_data.get('possibly_sensitive'),
            'possibly_sensitive_editable': raw_data.get('possibly_sensitive_editable'),
            'quote_count': raw_data.get('quote_count'),
            'entities': raw_data.get('entities'),
            'reply_count': raw_data.get('reply_count'),
            'favorite_count': raw_data.get('favorite_count'),
            'favorited': raw_data.get('favorited'),
            'retweet_count': raw_data.get('retweet_count')
        }
    }


def build_user_data(raw_data: dict) -> dict:
    return {
        **raw_data,
        'rest_id': raw_data['id'],
        'is_blue_verified': raw_data.get('ext_is_blue_verified'),
        'legacy': {
            'created_at': raw_data.get('created_at'),
            'name': raw_data.get('name'),
            'screen_name': raw_data.get('screen_name'),
            'profile_image_url_https': raw_data.get('profile_image_url_https'),
            'location': raw_data.get('location'),
            'description': raw_data.get('description'),
            'entities': raw_data.get('entities'),
            'pinned_tweet_ids_str': raw_data.get('pinned_tweet_ids_str'),
            'verified': raw_data.get('verified'),
            'possibly_sensitive': raw_data.get('possibly_sensitive'),
            'can_dm': raw_data.get('can_dm'),
            'can_media_tag': raw_data.get('can_media_tag'),
            'want_retweets': raw_data.get('want_retweets'),
            'default_profile': raw_data.get('default_profile'),
            'default_profile_image': raw_data.get('default_profile_image'),
            'has_custom_timelines': raw_data.get('has_custom_timelines'),
            'followers_count': raw_data.get('followers_count'),
            'fast_followers_count': raw_data.get('fast_followers_count'),
            'normal_followers_count': raw_data.get('normal_followers_count'),
            'friends_count': raw_data.get('friends_count'),
            'favourites_count': raw_data.get('favourites_count'),
            'listed_count': raw_data.get('listed_count'),
            'media_count': raw_data.get('media_count'),
            'statuses_count': raw_data.get('statuses_count'),
            'is_translator': raw_data.get('is_translator'),
            'translator_type': raw_data.get('translator_type'),
            'withheld_in_countries': raw_data.get('withheld_in_countries'),
            'url': raw_data.get('url'),
            'profile_banner_url': raw_data.get('profile_banner_url')
        }
    }


def flatten_params(params: dict) -> dict:
    flattened_params = {}
    for key, value in params.items():
        if isinstance(value, (list, dict)):
            value = json.dumps(value)
        flattened_params[key] = value
    return flattened_params


def b64_to_str(b64: str) -> str:
    return base64.b64decode(b64).decode()


def find_entry_by_type(entries, type_filter):
    for entry in entries:
        if entry.get('type') == type_filter:
            return entry
    return None


FILTERS = Literal[
    'media',
    'retweets',
    'native_video',
    'periscope',
    'vine',
    'images',
    'twimg',
    'links'
]


class SearchOptions(TypedDict):
    exact_phrases: list[str]
    or_keywords: list[str]
    exclude_keywords: list[str]
    hashtags: list[str]
    from_user: str
    to_user: str
    mentioned_users: list[str]
    filters: list[FILTERS]
    exclude_filters: list[FILTERS]
    urls: list[str]
    since: str
    until: str
    positive: bool
    negative: bool
    question: bool


def build_query(text: str, options: SearchOptions) -> str:
    """
    Builds a search query based on the given text and search options.

    Parameters
    ----------
    text : str
        The base text of the search query.
    options : SearchOptions
        A dictionary containing various search options.
        - exact_phrases: list[str]
            List of exact phrases to include in the search query.
        - or_keywords: list[str]
            List of keywords where tweets must contain at least
            one of these keywords.
        - exclude_keywords: list[str]
            A list of keywords that the tweet must contain these keywords.
        - hashtags: list[str]
            List of hashtags to include in the search query.
        - from_user: str
            Specify a username. Only tweets from this user will
            be includedin the search.
        - to_user: str
            Specify a username. Only tweets sent to this user will
            be included in the search.
        - mentioned_users: list[str]
            List of usernames. Only tweets mentioning these users will
            be included in the search.
        - filters: list[FILTERS]
            List of tweet filters to include in the search query.
        - exclude_filters: list[FILTERS]
            List of tweet filters to exclude from the search query.
        - urls: list[str]
            List of URLs. Only tweets containing these URLs will be
            included in the search.
        - since: str
            Specify a date (formatted as 'YYYY-MM-DD'). Only tweets since
            this date will be included in the search.
        - until: str
            Specify a date (formatted as 'YYYY-MM-DD'). Only tweets until
            this date will be included in the search.
        - positive: bool
            Include positive sentiment in the search.
        - negative: bool
            Include negative sentiment in the search.
        - question: bool
            Search for tweets in questionable form.

        https://developer.twitter.com/en/docs/twitter-api/v1/rules-and-filtering/search-operators

    Returns
    -------
    str
        The constructed Twitter search query.
    """
    if exact_phrases := options.get('exact_phrases'):
        text += ' ' + ' '.join(
            [f'"{i}"' for i in exact_phrases]
        )

    if or_keywords := options.get('or_keywords'):
        text += ' ' + ' OR '.join(or_keywords)

    if exclude_keywords := options.get('exclude_keywords'):
        text += ' ' + ' '.join(
            [f'-"{i}"' for i in exclude_keywords]
        )

    if hashtags := options.get('hashtags'):
        text += ' ' + ' '.join(
            [f'#{i}' for i in hashtags]
        )

    if from_user := options.get('from_user'):
        text +=f' from:{from_user}'

    if to_user := options.get('to_user'):
        text += f' to:{to_user}'

    if mentioned_users := options.get('mentioned_users'):
        text += ' ' + ' '.join(
            [f'@{i}' for i in mentioned_users]
        )

    if filters := options.get('filters'):
        text += ' ' + ' '.join(
            [f'filter:{i}' for i in filters]
        )

    if exclude_filters := options.get('exclude_filters'):
        text += ' ' + ' '.join(
            [f'-filter:{i}' for i in exclude_filters]
        )

    if urls := options.get('urls'):
        text += ' ' + ' '.join(
            [f'url:{i}' for i in urls]
        )

    if since := options.get('since'):
        text += f' since:{since}'

    if until := options.get('until'):
        text += f' until:{until}'

    if options.get('positive') is True:
        text += ' :)'

    if options.get('negative') is True:
        text += ' :('

    if options.get('question') is True:
        text += ' ?'

    return text



from __future__ import annotations

from time import sleep

import httpx

from .base import CaptchaSolver


class Capsolver(CaptchaSolver):
    """
    You can automatically unlock the account by passing the `captcha_solver`
    argument when initialising the :class:`.Client`.

    First, visit https://capsolver.com and obtain your Capsolver API key.
    Next, pass the Capsolver instance to the client as shown in the example.

    .. code-block:: python

        from twikit.twikit_async import Capsolver, Client
        solver = Capsolver(
            api_key='your_api_key',
            max_attempts=10
        )
        client = Client(captcha_solver=solver)

    Parameters
    ----------
    api_key : :class:`str`
        Capsolver API key.
    max_attempts : :class:`int`, default=3
        The maximum number of attempts to solve the captcha.
    get_result_interval : :class:`float`, default=1.0

    use_blob_data : :class:`bool`, default=False
    """

    def __init__(
        self,
        api_key: str,
        max_attempts: int = 3,
        get_result_interval: float = 1.0,
        use_blob_data: bool = False
    ) -> None:
        self.api_key = api_key
        self.get_result_interval = get_result_interval
        self.max_attempts = max_attempts
        self.use_blob_data = use_blob_data

    def create_task(self, task_data: dict) -> dict:
        data = {
            'clientKey': self.api_key,
            'task': task_data
        }
        response = httpx.post(
            'https://api.capsolver.com/createTask',
            json=data,
            headers={'content-type': 'application/json'}
        ).json()
        return response

    def get_task_result(self, task_id: str) -> dict:
        data = {
            'clientKey': self.api_key,
            'taskId': task_id
        }
        response = httpx.post(
            'https://api.capsolver.com/getTaskResult',
            json=data,
            headers={'content-type': 'application/json'}
        ).json()
        return response

    def solve_funcaptcha(self, blob: str) -> dict:
        if self.client.proxy is None:
            captcha_type = 'FunCaptchaTaskProxyLess'
        else:
            captcha_type = 'FunCaptchaTask'

        task_data = {
            'type': captcha_type,
            'websiteURL': 'https://iframe.arkoselabs.com',
            'websitePublicKey': self.CAPTCHA_SITE_KEY,
            'funcaptchaApiJSSubdomain': 'https://client-api.arkoselabs.com',
            'proxy': self.client.proxy
        }
        if self.use_blob_data:
            task_data['data'] = '{"blob":"%s"}' % blob
            task_data['userAgent'] = self.client._user_agent
        task = self.create_task(task_data)
        while True:
            sleep(self.get_result_interval)
            result = self.get_task_result(task['taskId'])
            if result['status'] in ('ready', 'failed'):
                return result



import re

from setuptools import setup

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

with open('./twikit/__init__.py') as f:
    version = re.findall(r"__version__ = '(.+)'", f.read())[0]


setup(
    name='twikit',
    version=version,
    install_requires=[
        'httpx[socks]',
        'filetype',
        'beautifulsoup4',
        'pyotp',
        'lxml',
        'webvtt-py',
        'm3u8',
        'Js2Py-3.13'
    ],
    python_requires='>=3.8',
    description='Twitter API wrapper for python with **no API key required**.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT',
    url='https://github.com/d60/twikit',
    package_data={'twikit': ['py.typed']}
)



from __future__ import annotations

from typing import TYPE_CHECKING, AsyncGenerator, NamedTuple

if TYPE_CHECKING:
    from .client.client import Client


class StreamingSession:
    """
    Represents a streaming session.

    Attributes
    ----------
    id : :class:`str`
        The ID or the session.
    topics : set[:class:`str`]
        The topics to stream.

    See Also
    --------
    .Client.get_streaming_session
    """
    def __init__(
        self, client: Client, session_id: str,
        stream: AsyncGenerator[Payload], topics: set[str], auto_reconnect: bool
    ) -> None:
        self._client = client
        self.id = session_id
        self._stream = stream
        self.topics = topics
        self.auto_reconnect = auto_reconnect

    async def reconnect(self) -> tuple[str, Payload]:
        """
        Reconnects the session.
        """
        stream = self._client._stream(self.topics)
        config_event = await anext(stream)
        self.id = config_event[1].config.session_id
        self._stream = stream
        return config_event

    async def update_subscriptions(
        self,
        subscribe: set[str] | None = None,
        unsubscribe: set[str] | None = None
    ) -> Payload:
        """
        Updates subscriptions for the session.

        Parameters
        ----------
        subscribe : set[:class:`str`], default=None
            Topics to subscribe to.
        unsubscribe : set[:class:`str`], default=None
            Topics to unsubscribe from.

        Examples
        --------
        >>> from twikit.streaming import Topic
        ...
        >>> subscribe_topics = {
        ...     Topic.tweet_engagement('1749528513'),
        ...     Topic.tweet_engagement('1765829534')
        ... }
        >>> unsubscribe_topics = {
        ...     Topic.tweet_engagement('17396176529'),
        ...     Topic.dm_update('17544932482-174455537996'),
        ...     Topic.dm_typing('17544932482-174455537996)'
        ... }
        >>> await session.update_subscriptions(
        ...     subscribe_topics, unsubscribe_topics
        ... )

        Note
        ----
        dm_update and dm_update cannot be added.

        See Also
        --------
        .Topic
        """
        return await self._client._update_subscriptions(
            self, subscribe, unsubscribe
        )

    async def __aiter__(self) -> AsyncGenerator[tuple[str, Payload]]:
        while True:
            async for event in self._stream:
                yield event
            if not self.auto_reconnect:
                break
            yield await self.reconnect()

    def __repr__(self) -> str:
        return f'<StreamingSession id="{self.id}">'


def _event_from_data(name: str, data: dict) -> StreamEventType:
    if name == 'config':
        session_id = data['session_id']
        subscription_ttl_millis = data['subscription_ttl_millis']
        heartbeat_millis = data['heartbeat_millis']
        return ConfigEvent(
            session_id, subscription_ttl_millis,
            heartbeat_millis
        )

    if name == 'subscriptions':
        errors = data['errors']
        return SubscriptionsEvent(errors)

    if name == 'tweet_engagement':
        like_count = data.get('like_count')
        retweet_count = data.get('retweet_count')
        quote_count = data.get('quote_count')
        reply_count = data.get('reply_count')
        view_count = None
        view_count_state = None
        if 'view_count_info' in data:
            view_count = data['view_count_info']['count']
            view_count_state = data['view_count_info']['state']
        return TweetEngagementEvent(
            like_count, retweet_count, view_count,
            view_count_state, quote_count, reply_count
        )

    if name == 'dm_update':
        conversation_id = data['conversation_id']
        user_id = data['user_id']
        return DMUpdateEvent(conversation_id, user_id)

    if name == 'dm_typing':
        conversation_id = data['conversation_id']
        user_id = data['user_id']
        return DMTypingEvent(conversation_id, user_id)


def _payload_from_data(data: dict) -> Payload:
    events = {
        name: _event_from_data(name, data)
        for (name, data) in data.items()
    }
    return Payload(**events)


class Payload(NamedTuple):
    """
    Represents a payload containing several types of events.
    """
    config: ConfigEvent | None = None  #: The configuration event.
    subscriptions: SubscriptionsEvent | None = None  #: The subscriptions event.
    tweet_engagement: TweetEngagementEvent | None = None  #: The tweet engagement event.
    dm_update: DMUpdateEvent | None = None  #: The direct message update event.
    dm_typing: DMTypingEvent | None = None  #: The direct message typing event.

    def __repr__(self) -> str:
        items = self._asdict().items()
        fields = [f'{i[0]}={i[1]}' for i in items if i[1] is not None]
        return f'Payload({" ".join(fields)})'


class ConfigEvent(NamedTuple):
    """
    Event representing configuration data.
    """
    session_id: str  #: The session ID associated with the configuration.
    subscription_ttl_millis: int  #: The time to live for the subscription.
    heartbeat_millis: int  #: The heartbeat interval in milliseconds.


class SubscriptionsEvent(NamedTuple):
    """
    Event representing subscription status.
    """
    errors: list  #: A list of errors.


class TweetEngagementEvent(NamedTuple):
    """
    Event representing tweet engagement metrics.
    """
    like_count: str | None  #: The number of likes on the tweet.
    retweet_count: str | None  #: The number of retweets of the tweet.
    view_count: str | None  #: The number of views of the tweet.
    view_count_state: str | None  #: The state of view count.
    quote_count: int | None  #: The number of quotes of the tweet.
    reply_count: int | None  # The number of Replies of the tweet.


class DMUpdateEvent(NamedTuple):
    """
    Event representing a (DM) update.
    """
    conversation_id: str  #: The ID of the conversation associated with the DM.
    user_id: str  #: ID of the user who sent the DM.


class DMTypingEvent(NamedTuple):
    """
    Event representing typing indication in a DM conversation.
    """
    conversation_id: str  #: The conversation where typing indication occurred.
    user_id: str  #: The ID of the typing user.

StreamEventType = (ConfigEvent | SubscriptionsEvent |
                   TweetEngagementEvent | DMTypingEvent | DMTypingEvent)


class Topic:
    """
    Utility class for generating topic strings for streaming.
    """
    @staticmethod
    def tweet_engagement(tweet_id: str) -> str:
        """
        Generates a topic string for tweet engagement events.

        Parameters
        ----------
        tweet_id : :class:`str`
            The ID of the tweet.

        Returns
        -------
        :class:`str`
            The topic string for tweet engagement events.
        """
        return f'/tweet_engagement/{tweet_id}'

    @staticmethod
    def dm_update(conversation_id: str) -> str:
        """
        Generates a topic string for direct message update events.

        Parameters
        ----------
        conversation_id : :class:`str`
            The ID of the conversation.
            Group ID (00000000) or partner_ID-your_ID (00000000-00000001)

        Returns
        -------
        :class:`str`
            The topic string for direct message update events.
        """
        return f'/dm_update/{conversation_id}'

    @staticmethod
    def dm_typing(conversation_id: str) -> str:
        """
        Generates a topic string for direct message typing events.

        Parameters
        ----------
        conversation_id : :class:`str`
            The ID of the conversation.
            Group ID (00000000) or partner_ID-your_ID (00000000-00000001)

        Returns
        -------
        :class:`str`
            The topic string for direct message typing events.
        """
        return f'/dm_typing/{conversation_id}'



import re
import bs4
import math
import time
import random
import base64
import hashlib
from typing import Union, List
from functools import reduce
from .cubic_curve import Cubic
from .interpolate import interpolate
from .rotation import convert_rotation_to_matrix
from .utils import float_to_hex, is_odd, base64_encode, handle_x_migration

ON_DEMAND_FILE_REGEX = re.compile(
    r"""['|\"]{1}ondemand\.s['|\"]{1}:\s*['|\"]{1}([\w]*)['|\"]{1}""", flags=(re.VERBOSE | re.MULTILINE))
INDICES_REGEX = re.compile(
    r"""(\(\w{1}\[(\d{1,2})\],\s*16\))+""", flags=(re.VERBOSE | re.MULTILINE))


class ClientTransaction:
    ADDITIONAL_RANDOM_NUMBER = 3
    DEFAULT_KEYWORD = "obfiowerehiring"
    DEFAULT_ROW_INDEX = None
    DEFAULT_KEY_BYTES_INDICES = None

    def __init__(self):
        self.home_page_response = None

    async def init(self, session, headers):
        home_page_response = await handle_x_migration(session, headers)

        self.home_page_response = self.validate_response(home_page_response)
        self.DEFAULT_ROW_INDEX, self.DEFAULT_KEY_BYTES_INDICES = await self.get_indices(
            self.home_page_response, session, headers)
        self.key = self.get_key(response=self.home_page_response)
        self.key_bytes = self.get_key_bytes(key=self.key)
        self.animation_key = self.get_animation_key(
            key_bytes=self.key_bytes, response=self.home_page_response)

    async def get_indices(self, home_page_response, session, headers):
        key_byte_indices = []
        response = self.validate_response(
            home_page_response) or self.home_page_response
        on_demand_file = ON_DEMAND_FILE_REGEX.search(str(response))
        if on_demand_file:
            on_demand_file_url = f"https://abs.twimg.com/responsive-web/client-web/ondemand.s.{on_demand_file.group(1)}a.js"
            on_demand_file_response = await session.request(method="GET", url=on_demand_file_url, headers=headers)
            key_byte_indices_match = INDICES_REGEX.finditer(
                str(on_demand_file_response.text))
            for item in key_byte_indices_match:
                key_byte_indices.append(item.group(2))
        if not key_byte_indices:
            raise Exception("Couldn't get KEY_BYTE indices")
        key_byte_indices = list(map(int, key_byte_indices))
        return key_byte_indices[0], key_byte_indices[1:]

    def validate_response(self, response: bs4.BeautifulSoup):
        if not isinstance(response, bs4.BeautifulSoup):
            raise Exception("invalid response")
        return response

    def get_key(self, response=None):
        response = self.validate_response(response) or self.home_page_response
        # <meta name="twitter-site-verification" content="mentU...+1yPz..../IcNS+......./RaF...R+b"/>
        element = response.select_one("[name='twitter-site-verification']")
        if not element:
            raise Exception("Couldn't get key from the page source")
        return element.get("content")

    def get_key_bytes(self, key: str):
        return list(base64.b64decode(bytes(key, 'utf-8')))

    def get_frames(self, response=None):
        # loading-x-anim-0...loading-x-anim-3
        response = self.validate_response(response) or self.home_page_response
        return response.select("[id^='loading-x-anim']")

    def get_2d_array(self, key_bytes: List[Union[float, int]], response, frames: bs4.ResultSet = None):
        if not frames:
            frames = self.get_frames(response)
        # return list(list(frames[key[5] % 4].children)[0].children)[1].get("d")[9:].split("C")
        return [[int(x) for x in re.sub(r"[^\d]+", " ", item).strip().split()] for item in list(list(frames[key_bytes[5] % 4].children)[0].children)[1].get("d")[9:].split("C")]

    def solve(self, value, min_val, max_val, rounding: bool):
        result = value * (max_val-min_val) / 255 + min_val
        return math.floor(result) if rounding else round(result, 2)

    def animate(self, frames, target_time):
        # from_color = f"#{''.join(['{:x}'.format(digit) for digit in frames[:3]])}"
        # to_color = f"#{''.join(['{:x}'.format(digit) for digit in frames[3:6]])}"
        # from_rotation = "rotate(0deg)"
        # to_rotation = f"rotate({solve(frames[6], 60, 360, True)}deg)"
        # easing_values = [solve(value, -1 if count % 2 else 0, 1, False)
        #                  for count, value in enumerate(frames[7:])]
        # easing = f"cubic-bezier({','.join([str(value) for value in easing_values])})"
        # current_time = round(target_time / 10) * 10

        from_color = [float(item) for item in [*frames[:3], 1]]
        to_color = [float(item) for item in [*frames[3:6], 1]]
        from_rotation = [0.0]
        to_rotation = [self.solve(float(frames[6]), 60.0, 360.0, True)]
        frames = frames[7:]
        curves = [self.solve(float(item), is_odd(counter), 1.0, False)
                  for counter, item in enumerate(frames)]
        cubic = Cubic(curves)
        val = cubic.get_value(target_time)
        color = interpolate(from_color, to_color, val)
        color = [value if value > 0 else 0 for value in color]
        rotation = interpolate(from_rotation, to_rotation, val)
        matrix = convert_rotation_to_matrix(rotation[0])
        # str_arr = [format(int(round(color[i])), '02x') for i in range(len(color) - 1)]
        # str_arr = [format(int(round(color[i])), 'x') for i in range(len(color) - 1)]
        str_arr = [format(round(value), 'x') for value in color[:-1]]
        for value in matrix:
            rounded = round(value, 2)
            if rounded < 0:
                rounded = -rounded
            hex_value = float_to_hex(rounded)
            str_arr.append(f"0{hex_value}".lower() if hex_value.startswith(
                ".") else hex_value if hex_value else '0')
        str_arr.extend(["0", "0"])
        animation_key = re.sub(r"[.-]", "", "".join(str_arr))
        return animation_key

    def get_animation_key(self, key_bytes, response):
        total_time = 4096
        # row_index, frame_time = [key_bytes[2] % 16, key_bytes[12] % 16 * (key_bytes[14] % 16) * (key_bytes[7] % 16)]
        # row_index, frame_time = [key_bytes[2] % 16, key_bytes[2] % 16 * (key_bytes[42] % 16) * (key_bytes[45] % 16)]

        row_index = key_bytes[self.DEFAULT_ROW_INDEX] % 16
        frame_time = reduce(lambda num1, num2: num1*num2,
                            [key_bytes[index] % 16 for index in self.DEFAULT_KEY_BYTES_INDICES])
        arr = self.get_2d_array(key_bytes, response)
        frame_row = arr[row_index]

        target_time = float(frame_time) / total_time
        animation_key = self.animate(frame_row, target_time)
        return animation_key

    def generate_transaction_id(self, method: str, path: str, response=None, key=None, animation_key=None, time_now=None):
        time_now = time_now or math.floor(
            (time.time() * 1000 - 1682924400 * 1000) / 1000)
        time_now_bytes = [(time_now >> (i * 8)) & 0xFF for i in range(4)]
        key = key or self.key or self.get_key(response)
        key_bytes = self.get_key_bytes(key)
        animation_key = animation_key or self.animation_key or self.get_animation_key(
            key_bytes, response)
        # hash_val = hashlib.sha256(f"{method}!{path}!{time_now}bird{animation_key}".encode()).digest()
        hash_val = hashlib.sha256(
            f"{method}!{path}!{time_now}{self.DEFAULT_KEYWORD}{animation_key}".encode()).digest()
        # hash_bytes = [int(hash_val[i]) for i in range(len(hash_val))]
        hash_bytes = list(hash_val)
        random_num = random.randint(0, 255)
        bytes_arr = [*key_bytes, *time_now_bytes, *
                     hash_bytes[:16], self.ADDITIONAL_RANDOM_NUMBER]
        out = bytearray(
            [random_num, *[item ^ random_num for item in bytes_arr]])
        return base64_encode(out).strip("=")



from __future__ import annotations

from typing import TypedDict, TYPE_CHECKING

if TYPE_CHECKING:
    from .client.client import Client


class Trend:
    """
    Attributes
    ----------
    name : :class:`str`
        The name of the trending topic.
    tweets_count : :class:`int`
        The count of tweets associated with the trend.
    domain_context : :class:`str`
        The context or domain associated with the trend.
    grouped_trends : :class:`list`[:class:`str`]
        A list of trend names grouped under the main trend.
    """

    def __init__(self, client: Client, data: dict) -> None:
        self._client = client

        metadata: dict = data['trendMetadata']
        self.name: str = data['name']
        self.tweets_count: int | None = metadata.get('metaDescription')
        self.domain_context: str = metadata.get('domainContext')
        self.grouped_trends: list[str] = [
            trend['name'] for trend in data.get('groupedTrends', [])
        ]

    def __repr__(self) -> str:
        return f'<Trend name="{self.name}">'


class PlaceTrends(TypedDict):
    trends: list[PlaceTrend]
    as_of: str
    created_at: str
    locations: dict


class PlaceTrend:
    """
    Attributes
    ----------
    name : :class:`str`
        The name of the trend.
    url : :class:`str`
        The URL to view the trend.
    query : :class:`str`
        The search query corresponding to the trend.
    tweet_volume : :class:`int`
        The volume of tweets associated with the trend.
    """
    def __init__(self, client: Client, data: dict) -> None:
        self._client = client

        self.name: str = data['name']
        self.url: str = data['url']
        self.promoted_content: None = data['promoted_content']
        self.query: str = data['query']
        self.tweet_volume: int = data['tweet_volume']

    def __repr__(self) -> str:
        return f'<PlaceTrend name="{self.name}">'


class Location:
    def __init__(self, client: Client, data: dict) -> None:
        self._client = client

        self.woeid: int = data['woeid']
        self.country: str = data['country']
        self.country_code: str = data['countryCode']
        self.name: str = data['name']
        self.parentid: int = data['parentid']
        self.placeType: dict = data['placeType']
        self.url: str = data['url']

    async def get_trends(self) -> PlaceTrends:
        return await self._client.get_place_trends(self.woeid)

    def __repr__(self) -> str:
        return f'<Location name="{self.name}" woeid={self.woeid}>'

    def __eq__(self, __value: object) -> bool:
        return isinstance(__value, Location) and self.woeid == __value.woeid

    def __ne__(self, __value: object) -> bool:
        return not self == __value


from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Literal

from ..utils import Result, timestamp_to_datetime

if TYPE_CHECKING:
    from .client import GuestClient
    from .tweet import Tweet


class User:
    """
    Attributes
    ----------
    id : :class:`str`
        The unique identifier of the user.
    created_at : :class:`str`
        The date and time when the user account was created.
    name : :class:`str`
        The user's name.
    screen_name : :class:`str`
        The user's screen name.
    profile_image_url : :class:`str`
        The URL of the user's profile image (HTTPS version).
    profile_banner_url : :class:`str`
        The URL of the user's profile banner.
    url : :class:`str`
        The user's URL.
    location : :class:`str`
        The user's location information.
    description : :class:`str`
        The user's profile description.
    description_urls : :class:`list`
        URLs found in the user's profile description.
    urls : :class:`list`
        URLs associated with the user.
    pinned_tweet_ids : :class:`str`
        The IDs of tweets that the user has pinned to their profile.
    is_blue_verified : :class:`bool`
        Indicates if the user is verified with a blue checkmark.
    verified : :class:`bool`
        Indicates if the user is verified.
    possibly_sensitive : :class:`bool`
        Indicates if the user's content may be sensitive.
    can_media_tag : :class:`bool`
        Indicates whether the user can be tagged in media.
    want_retweets : :class:`bool`
        Indicates if the user wants retweets.
    default_profile : :class:`bool`
        Indicates if the user has the default profile.
    default_profile_image : :class:`bool`
        Indicates if the user has the default profile image.
    has_custom_timelines : :class:`bool`
        Indicates if the user has custom timelines.
    followers_count : :class:`int`
        The count of followers.
    fast_followers_count : :class:`int`
        The count of fast followers.
    normal_followers_count : :class:`int`
        The count of normal followers.
    following_count : :class:`int`
        The count of users the user is following.
    favourites_count : :class:`int`
        The count of favorites or likes.
    listed_count : :class:`int`
        The count of lists the user is a member of.
    media_count : :class:`int`
        The count of media items associated with the user.
    statuses_count : :class:`int`
        The count of tweets.
    is_translator : :class:`bool`
        Indicates if the user is a translator.
    translator_type : :class:`str`
        The type of translator.
    profile_interstitial_type : :class:`str`
        The type of profile interstitial.
    withheld_in_countries : list[:class:`str`]
        Countries where the user's content is withheld.
    """

    def __init__(self, client: GuestClient, data: dict) -> None:
        self._client = client
        legacy = data['legacy']

        self.id: str = data['rest_id']
        self.created_at: str = legacy['created_at']
        self.name: str = legacy['name']
        self.screen_name: str = legacy['screen_name']
        self.profile_image_url: str = legacy['profile_image_url_https']
        self.profile_banner_url: str = legacy.get('profile_banner_url')
        self.url: str = legacy.get('url')
        self.location: str = legacy['location']
        self.description: str = legacy['description']
        self.description_urls: list = legacy['entities']['description']['urls']
        self.urls: list = legacy['entities'].get('url', {}).get('urls')
        self.pinned_tweet_ids: list[str] = legacy['pinned_tweet_ids_str']
        self.is_blue_verified: bool = data['is_blue_verified']
        self.verified: bool = legacy['verified']
        self.possibly_sensitive: bool = legacy['possibly_sensitive']
        self.default_profile: bool = legacy['default_profile']
        self.default_profile_image: bool = legacy['default_profile_image']
        self.has_custom_timelines: bool = legacy['has_custom_timelines']
        self.followers_count: int = legacy['followers_count']
        self.fast_followers_count: int = legacy['fast_followers_count']
        self.normal_followers_count: int = legacy['normal_followers_count']
        self.following_count: int = legacy['friends_count']
        self.favourites_count: int = legacy['favourites_count']
        self.listed_count: int = legacy['listed_count']
        self.media_count = legacy['media_count']
        self.statuses_count: int = legacy['statuses_count']
        self.is_translator: bool = legacy['is_translator']
        self.translator_type: str = legacy['translator_type']
        self.withheld_in_countries: list[str] = legacy['withheld_in_countries']
        self.protected: bool = legacy.get('protected', False)

    @property
    def created_at_datetime(self) -> datetime:
        return timestamp_to_datetime(self.created_at)

    async def get_tweets(self, tweet_type: Literal['Tweets'] = 'Tweets', count: int = 40) -> list[Tweet]:
        """
        Retrieves the user's tweets.

        Parameters
        ----------
        tweet_type : {'Tweets'}, default='Tweets'
            The type of tweets to retrieve.
        count : :class:`int`, default=40
            The number of tweets to retrieve.

        Returns
        -------
        list[:class:`.tweet.Tweet`]
            A list of `Tweet` objects.

        Examples
        --------
        >>> user = await client.get_user_by_screen_name('example_user')
        >>> tweets = await user.get_tweets()
        >>> for tweet in tweets:
        ...    print(tweet)
        <Tweet id="...">
        <Tweet id="...">
        ...
        ...
        """
        return await self._client.get_user_tweets(self.id, tweet_type, count)

    async def get_highlights_tweets(self, count: int = 20, cursor: str | None = None) -> Result[Tweet]:
        """
        Retrieves highlighted tweets from the user's timeline.

        Parameters
        ----------
        count : :class:`int`, default=20
            The number of tweets to retrieve.

        Returns
        -------
        Result[:class:`.tweet.Tweet`]
            An instance of the `Result` class containing the highlighted tweets.

        Examples
        --------
        >>> result = await user.get_highlights_tweets()
        >>> for tweet in result:
        ...     print(tweet)
        <Tweet id="...">
        <Tweet id="...">
        ...
        ...

        >>> more_results = await result.next()  # Retrieve more highlighted tweets
        >>> for tweet in more_results:
        ...     print(tweet)
        <Tweet id="...">
        <Tweet id="...">
        ...
        ...
        """
        return await self._client.get_user_highlights_tweets(self.id, count, cursor)

    async def update(self) -> None:
        new = await self._client.get_user_by_id(self.id)
        self.__dict__.update(new.__dict__)

    def __repr__(self) -> str:
        return f'<User id="{self.id}">'

    def __eq__(self, __value: object) -> bool:
        return isinstance(__value, User) and self.id == __value.id

    def __ne__(self, __value: object) -> bool:
        return not self == __value
